# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
"""
=================================================================================
Ensemble Similarity Calculations --- :mod:`MDAnalysis.analysis.encore.similarity`
=================================================================================

:Author: Matteo Tiberti, Wouter Boomsma, Tone Bengtsen

.. versionadded:: 0.16.0

.. deprecated:: 2.8.0
   This module is deprecated in favour of the 
   MDAKit `mdaencore <https://mdanalysis.org/mdaencore/>`_ and will be removed
   in MDAnalysis 3.0.0.

The module contains implementations of similarity measures between protein
ensembles described in :footcite:p:`LindorffLarsen2009`. The implementation and
examples are described in :footcite:p:`Tiberti2015`.

The module includes facilities for handling ensembles and trajectories through
the :class:`Universe` class, performing clustering or dimensionality reduction
of the ensemble space, estimating multivariate probability distributions from
the input data, and more. ENCORE can be used to compare experimental and
simulation-derived ensembles, as well as estimate the convergence of
trajectories from time-dependent simulations.

ENCORE includes three different methods for calculations of similarity measures
between ensembles implemented in individual functions:

+ **Harmonic Ensemble Similarity** : :func:`hes`
+ **Clustering Ensemble Similarity** : :func:`ces`
+ **Dimensional Reduction Ensemble Similarity** : :func:`dres`

as well as two methods to evaluate the convergence of trajectories:

+ **Clustering based convergence evaluation** : :func:`ces_convergence`
+ **Dimensionality-reduction based convergence evaluation** : :func:`dres_convergence`

When using this module in published work please cite :footcite:p:`Tiberti2015`.

.. rubric:: References

.. footbibliography::

.. _Examples:
Examples
========

The examples show how to use ENCORE to calculate a similarity measurement
of two simple ensembles. The ensembles are obtained from the MDAnalysis
test suite for two different simulations of the protein AdK.

To calculate the Harmonic Ensemble Similarity (:func:`hes`)
two ensemble objects are first created and then used for calculation:

    >>> from MDAnalysis import Universe
    >>> import MDAnalysis.analysis.encore as encore
    >>> from MDAnalysis.tests.datafiles import PSF, DCD, DCD2
    >>> ens1 = Universe(PSF, DCD)
    >>> ens2 = Universe(PSF, DCD2)
    >>> HES, details = encore.hes([ens1, ens2])
    >>> print(HES)
    [[       0.         38279540.04524205]
     [38279540.04524205        0.        ]]

HES can assume any non-negative value, i.e. no upper bound exists and the
measurement can therefore be used as an absolute scale.

The calculation of the Clustering Ensemble Similarity (:func:`ces`)
is computationally more expensive. It is based on clustering algorithms that in
turn require a similarity matrix between the frames the ensembles are made
of. The similarity matrix is derived from a distance matrix (By default a RMSD
matrix; a full RMSD matrix between each pairs of elements needs to be computed).
The RMSD matrix is automatically calculated:

    >>> from MDAnalysis import Universe
    >>> import MDAnalysis.analysis.encore as encore
    >>> from MDAnalysis.tests.datafiles import PSF, DCD, DCD2
    >>> ens1 = Universe(PSF, DCD)
    >>> ens2 = Universe(PSF, DCD2)
    >>> CES, details = encore.ces([ens1, ens2])
    >>> print(CES)
    [[0.         0.68070702]
     [0.68070702 0.        ]]

The RMSD matrix can also be separately calculated to reuse it, e.g. for running
CES with different parameters or running the
Dimensional Reduction Ensemble Similarity (:func:`dres`) method.
DRES is based on the estimation of the probability density in
a dimensionally-reduced conformational space of the ensembles, obtained from
the original space using either the Stochastic Proximity Embedding algorithm or
the Principal Component Analysis.
In the following example the dimensions are reduced to 3 using the
RMSD matrix and the default SPE dimensional reduction method:

    >>> from MDAnalysis import Universe
    >>> import MDAnalysis.analysis.encore as encore
    >>> from MDAnalysis.tests.datafiles import PSF, DCD, DCD2
    >>> ens1 = Universe(PSF, DCD)
    >>> ens2 = Universe(PSF, DCD2)
    >>> rmsd_matrix = encore.get_distance_matrix(
    ...                             encore.utils.merge_universes([ens1, ens2]))
    >>> DRES,details = encore.dres([ens1, ens2],
    ...                             distance_matrix = rmsd_matrix)

The RMSD matrix can also be saved on disk with the option ``save_matrix``: ::

    rmsd_matrix = encore.get_distance_matrix(
                                    encore.utils.merge_universes([ens1, ens2]),
                                    save_matrix="rmsd.npz")

It can then be loaded and reused at a later time instead of being recalculated: ::

    rmsd_matrix = encore.get_distance_matrix(
                                    encore.utils.merge_universes([ens1, ens2]),
                                    load_matrix="rmsd.npz")

In addition to the quantitative similarity estimate, the dimensional reduction
can easily be visualized, see the ``Example`` section in
:mod:`MDAnalysis.analysis.encore.dimensionality_reduction.reduce_dimensionality`.
Due to the stochastic nature of SPE, two identical ensembles will not
necessarily result in an exactly 0 estimate of the similarity, but will be very
close. For the same reason, calculating the similarity with the :func:`dres`
twice will not result in necessarily identical values but rather two very close
values.

It should be noted that both in :func:`ces` and :func:`dres` the similarity is
evaluated using the Jensen-Shannon divergence resulting in an upper bound of
ln(2), which indicates no similarity between the ensembles and a lower bound
of 0.0 signifying two identical ensembles. In contrast, the :func:`hes` function uses
a symmetrized version of the Kullback-Leibler divergence, which is unbounded.


Functions for ensemble comparisons
==================================

.. autofunction:: hes
   :noindex:

.. autofunction:: ces
   :noindex:

.. autofunction:: dres
   :noindex:

Function reference
==================

.. All functions are included via automodule :members:.

"""
import warnings
import logging

import numpy as np
import scipy.stats

import MDAnalysis as mda

from ...coordinates.memory import MemoryReader
from .confdistmatrix import get_distance_matrix
from .bootstrap import (
    get_distance_matrix_bootstrap_samples,
    get_ensemble_bootstrap_samples,
)
from .clustering.cluster import cluster
from .clustering.ClusteringMethod import AffinityPropagationNative
from .dimensionality_reduction.DimensionalityReductionMethod import (
    StochasticProximityEmbeddingNative,
)
from .dimensionality_reduction.reduce_dimensionality import (
    reduce_dimensionality,
)
from .covariance import (
    covariance_matrix,
    ml_covariance_estimator,
    shrinkage_covariance_estimator,
)
from .utils import merge_universes
from .utils import trm_indices_diag, trm_indices_nodiag

# Low boundary value for log() argument - ensure no nans
EPSILON = 1e-15

xlogy = np.vectorize(
    lambda x, y: 0.0 if (x <= EPSILON and y <= EPSILON) else x * np.log(y)
)


def discrete_kullback_leibler_divergence(pA, pB):
    """Kullback-Leibler divergence between discrete probability distribution.
    Notice that since this measure is not symmetric ::
    :math:`d_{KL}(p_A,p_B) != d_{KL}(p_B,p_A)`

    Parameters
    ----------

    pA : iterable of floats
        First discrete probability density function

    pB : iterable of floats
        Second discrete probability density function

    Returns
    -------

    dkl : float
        Discrete Kullback-Liebler divergence
    """

    return np.sum(xlogy(pA, pA / pB))


# discrete dJS
def discrete_jensen_shannon_divergence(pA, pB):
    """Jensen-Shannon divergence between discrete probability distributions.

    Parameters
    ----------

    pA : iterable of floats
        First discrete probability density function

    pB : iterable of floats
        Second discrete probability density function

    Returns
    -------

    djs : float
        Discrete Jensen-Shannon divergence
    """
    return 0.5 * (
        discrete_kullback_leibler_divergence(pA, (pA + pB) * 0.5)
        + discrete_kullback_leibler_divergence(pB, (pA + pB) * 0.5)
    )


# calculate harmonic similarity
def harmonic_ensemble_similarity(sigma1, sigma2, x1, x2):
    """
    Calculate the harmonic ensemble similarity measure
    as defined in :footcite:p:`Tiberti2015`.

    Parameters
    ----------

    sigma1 : numpy.array
        Covariance matrix for the first ensemble.

    sigma2 : numpy.array
        Covariance matrix for the second ensemble.

    x1: numpy.array
        Mean for the estimated normal multivariate distribution of the first
        ensemble.

    x2: numpy.array
        Mean for the estimated normal multivariate distribution of the second
        ensemble.

    Returns
    -------

        dhes : float
            harmonic similarity measure
    """

    # Inverse covariance matrices
    sigma1_inv = np.linalg.pinv(sigma1)
    sigma2_inv = np.linalg.pinv(sigma2)

    # Difference between average vectors
    d_avg = x1 - x2

    # Distance measure
    trace = np.trace(
        np.dot(sigma1, sigma2_inv)
        + np.dot(sigma2, sigma1_inv)
        - 2 * np.identity(sigma1.shape[0])
    )

    d_hes = 0.25 * (
        np.dot(np.transpose(d_avg), np.dot(sigma1_inv + sigma2_inv, d_avg))
        + trace
    )
    return d_hes


def clustering_ensemble_similarity(
    cc, ens1, ens1_id, ens2, ens2_id, select="name CA"
):
    """Clustering ensemble similarity: calculate the probability densities from
     the clusters and calculate discrete Jensen-Shannon divergence.

    Parameters
    ----------

    cc : encore.clustering.ClustersCollection
        Collection from cluster calculated by a clustering algorithm
        (e.g. Affinity propagation)

    ens1 : :class:`~MDAnalysis.core.universe.Universe`
        First ensemble to be used in comparison

    ens1_id : int
        First ensemble id as detailed in the ClustersCollection metadata

    ens2 : :class:`~MDAnalysis.core.universe.Universe`
        Second ensemble to be used in comparison

    ens2_id : int
        Second ensemble id as detailed in the ClustersCollection metadata

    select : str
        Atom selection string in the MDAnalysis format. Default is "name CA".

    Returns
    -------

    djs : float
        Jensen-Shannon divergence between the two ensembles, as calculated by
        the clustering ensemble similarity method
    """
    ens1_coordinates = ens1.trajectory.timeseries(
        ens1.select_atoms(select), order="fac"
    )
    ens2_coordinates = ens2.trajectory.timeseries(
        ens2.select_atoms(select), order="fac"
    )
    tmpA = np.array(
        [
            np.where(c.metadata["ensemble_membership"] == ens1_id)[0].shape[0]
            / float(ens1_coordinates.shape[0])
            for c in cc
        ]
    )
    tmpB = np.array(
        [
            np.where(c.metadata["ensemble_membership"] == ens2_id)[0].shape[0]
            / float(ens2_coordinates.shape[0])
            for c in cc
        ]
    )

    # Exclude clusters which have 0 elements in both ensembles
    pA = tmpA[tmpA + tmpB > EPSILON]
    pB = tmpB[tmpA + tmpB > EPSILON]

    return discrete_jensen_shannon_divergence(pA, pB)


def cumulative_clustering_ensemble_similarity(
    cc, ens1_id, ens2_id, ens1_id_min=1, ens2_id_min=1
):
    """
    Calculate clustering ensemble similarity between joined ensembles.
    This means that, after clustering has been performed, some ensembles are
    merged and the dJS is calculated between the probability distributions of
    the two clusters groups. In particular, the two ensemble groups are defined
    by their ensembles id: one of the two joined ensembles will comprise all
    the ensembles with id [ens1_id_min, ens1_id], and the other ensembles will
    comprise all the ensembles with id [ens2_id_min, ens2_id].

    Parameters
    ----------

    cc : encore.ClustersCollection
            Collection from cluster calculated by a clustering algorithm
            (e.g. Affinity propagation)

    ens1_id : int
            First ensemble id as detailed in the ClustersCollection
            metadata

    ens2_id : int
            Second ensemble id as detailed in the ClustersCollection
            metadata

    Returns
    -------

    djs : float
            Jensen-Shannon divergence between the two ensembles, as
            calculated by the clustering ensemble similarity method

    """

    ensA = [
        np.where(
            np.logical_and(
                c.metadata["ensemble_membership"] <= ens1_id,
                c.metadata["ensemble_membership"],
            )
            >= ens1_id_min
        )[0].shape[0]
        for c in cc
    ]
    ensB = [
        np.where(
            np.logical_and(
                c.metadata["ensemble_membership"] <= ens2_id,
                c.metadata["ensemble_membership"],
            )
            >= ens2_id_min
        )[0].shape[0]
        for c in cc
    ]
    sizeA = float(np.sum(ensA))
    sizeB = float(np.sum(ensB))

    tmpA = np.array(ensA) / sizeA
    tmpB = np.array(ensB) / sizeB

    # Exclude clusters which have 0 elements in both ensembles
    pA = tmpA[tmpA + tmpB > EPSILON]
    pB = tmpB[tmpA + tmpB > EPSILON]

    return discrete_jensen_shannon_divergence(pA, pB)


def gen_kde_pdfs(embedded_space, ensemble_assignment, nensembles, nsamples):
    """
    Generate Kernel Density Estimates (KDE) from embedded spaces and
    elaborate the coordinates for later use.

    Parameters
    ----------

    embedded_space : numpy.array
        Array containing the coordinates of the embedded space

    ensemble_assignment : numpy.array
        Array containing one int per ensemble conformation. These allow to
        distinguish, in the complete embedded space, which conformations
        belong to each ensemble. For instance if ensemble_assignment
        is [1,1,1,1,2,2], it means that the first four conformations belong
        to ensemble 1 and the last two to ensemble 2

    nensembles : int
        Number of ensembles

    nsamples : int
        samples to be drawn from the ensembles. Will be required in
        a later stage in order to calculate dJS.

    Returns
    -------

    kdes : scipy.stats.gaussian_kde
        KDEs calculated from ensembles

    resamples : list of numpy.array
        For each KDE, draw samples according to the probability distribution
        of the KDE mixture model

    embedded_ensembles : list of numpy.array
        List of numpy.array containing, each one, the elements of the
        embedded space belonging to a certain ensemble
    """
    kdes = []
    embedded_ensembles = []
    resamples = []

    for i in range(1, nensembles + 1):
        this_embedded = embedded_space.transpose()[
            np.where(np.array(ensemble_assignment) == i)
        ].transpose()
        embedded_ensembles.append(this_embedded)
        kdes.append(scipy.stats.gaussian_kde(this_embedded))

    # # Set number of samples
    # if not nsamples:
    #     nsamples = this_embedded.shape[1] * 10

    # Resample according to probability distributions
    for this_kde in kdes:
        resamples.append(this_kde.resample(nsamples))

    return (kdes, resamples, embedded_ensembles)


def dimred_ensemble_similarity(
    kde1,
    resamples1,
    kde2,
    resamples2,
    ln_P1_exp_P1=None,
    ln_P2_exp_P2=None,
    ln_P1P2_exp_P1=None,
    ln_P1P2_exp_P2=None,
):
    r"""Calculate the Jensen-Shannon divergence according the Dimensionality
    reduction method.

    In this case, we have continuous probability densities, this we need to
    integrate over the measurable space. The aim is to first calculate the
    Kullback-Liebler divergence, which is defined as:

    .. math::

       D_{KL}(P(x) || Q(x)) =
           \int_{-\infty}^{\infty}P(x_i) ln(P(x_i)/Q(x_i)) =
           \langle{}ln(P(x))\rangle{}_P - \langle{}ln(Q(x))\rangle{}_P

    where the :math:`\langle{}.\rangle{}_P` denotes an expectation calculated
    under the distribution P. We can, thus, just estimate the expectation
    values of the components to get an estimate of dKL.  Since the
    Jensen-Shannon distance is actually more complex, we need to estimate four
    expectation values:

    .. math::
         \langle{}log(P(x))\rangle{}_P

         \langle{}log(Q(x))\rangle{}_Q

         \langle{}log(0.5*(P(x)+Q(x)))\rangle{}_P

         \langle{}log(0.5*(P(x)+Q(x)))\rangle{}_Q

    Parameters
    ----------

    kde1 : scipy.stats.gaussian_kde
        Kernel density estimation for ensemble 1

    resamples1 : numpy.array
        Samples drawn according do kde1. Will be used as samples to
        calculate the expected values according to 'P' as detailed before.

    kde2 : scipy.stats.gaussian_kde
            Kernel density estimation for ensemble 2

    resamples2 : numpy.array
        Samples drawn according do kde2. Will be used as sample to
        calculate the expected values according to 'Q' as detailed before.

    ln_P1_exp_P1 : float or None
        Use this value for :math:`\langle{}log(P(x))\rangle{}_P`; if ``None``,
        calculate it instead

    ln_P2_exp_P2 : float or None
        Use this value for :math:`\langle{}log(Q(x))\rangle{}_Q`; if
        ``None``, calculate it instead

    ln_P1P2_exp_P1 : float or None
        Use this value for
        :math:`\langle{}log(0.5*(P(x)+Q(x)))\rangle{}_P`;
        if ``None``, calculate it instead

    ln_P1P2_exp_P2 : float or None
        Use this value for
        :math:`\langle{}log(0.5*(P(x)+Q(x)))\rangle{}_Q`;
        if ``None``, calculate it instead

    Returns
    -------
    djs : float
        Jensen-Shannon divergence calculated according to the dimensionality
        reduction method

    """

    if (
        not ln_P1_exp_P1
        and not ln_P2_exp_P2
        and not ln_P1P2_exp_P1
        and not ln_P1P2_exp_P2
    ):
        ln_P1_exp_P1 = np.average(np.log(kde1.evaluate(resamples1)))
        ln_P2_exp_P2 = np.average(np.log(kde2.evaluate(resamples2)))
        ln_P1P2_exp_P1 = np.average(
            np.log(
                0.5 * (kde1.evaluate(resamples1) + kde2.evaluate(resamples1))
            )
        )
        ln_P1P2_exp_P2 = np.average(
            np.log(
                0.5 * (kde1.evaluate(resamples2) + kde2.evaluate(resamples2))
            )
        )

    return 0.5 * (
        ln_P1_exp_P1 - ln_P1P2_exp_P1 + ln_P2_exp_P2 - ln_P1P2_exp_P2
    )


def cumulative_gen_kde_pdfs(
    embedded_space,
    ensemble_assignment,
    nensembles,
    nsamples,
    ens_id_min=1,
    ens_id_max=None,
):
    """
    Generate Kernel Density Estimates (KDE) from embedded spaces and
    elaborate the coordinates for later use. However, consider more than
    one ensemble as the space on which the KDE will be generated. In
    particular, will use ensembles with ID [ens_id_min, ens_id_max].


    Parameters
    ----------

    embedded_space : numpy.array
            Array containing the coordinates of the embedded space

    ensemble_assignment : numpy.array
            array containing one int per ensemble conformation. These allow
            to distinguish, in the complete embedded space, which
            conformations belong to each ensemble. For instance if
            ensemble_assignment is [1,1,1,1,2,2], it means that the first
            four conformations belong to ensemble 1 and the last two
            to ensemble 2

    nensembles : int
            Number of ensembles

    nsamples : int
        Samples to be drawn from the ensembles. Will be required in a later
        stage in order to calculate dJS.

    ens_id_min : int
        Minimum ID of the ensemble to be considered; see description

    ens_id_max : int
        Maximum ID of the ensemble to be considered; see description. If None,
        it will be set to the maximum possible value given the number of
        ensembles.

    Returns
    -------

    kdes : scipy.stats.gaussian_kde
            KDEs calculated from ensembles

    resamples : list of numpy.array
            For each KDE, draw samples according to the probability
            distribution of the kde mixture model

    embedded_ensembles : list of numpy.array
            List of numpy.array containing, each one, the elements of the
            embedded space belonging to a certain ensemble


    """

    kdes = []
    embedded_ensembles = []
    resamples = []
    if not ens_id_max:
        ens_id_max = nensembles + 1
    for i in range(ens_id_min, ens_id_max):
        this_embedded = embedded_space.transpose()[
            np.where(
                np.logical_and(
                    ensemble_assignment >= ens_id_min, ensemble_assignment <= i
                )
            )
        ].transpose()
        embedded_ensembles.append(this_embedded)
        kdes.append(scipy.stats.gaussian_kde(this_embedded))

    # Resample according to probability distributions
    for this_kde in kdes:
        resamples.append(this_kde.resample(nsamples))

    return (kdes, resamples, embedded_ensembles)


def write_output(
    matrix, base_fname=None, header="", suffix="", extension="dat"
):
    """
    Write output matrix with a nice format, to stdout and optionally a file.

    Parameters
    ----------

    matrix : encore.utils.TriangularMatrix
        Matrix containing the values to be printed

    base_fname : str
        Basic filename for output. If None, no files will be written, and
        the matrix will be just printed on standard output

    header : str
        Text to be written just before the matrix

    suffix : str
        String to be concatenated to basename, in order to get the final
        file name

    extension : str
        Extension for the output file

    """

    if base_fname is not None:
        fname = base_fname + "-" + suffix + "." + extension
    else:
        fname = None
    matrix.square_print(header=header, fname=fname)


def prepare_ensembles_for_convergence_increasing_window(
    ensemble, window_size, select="name CA"
):
    """
    Generate ensembles to be fed to ces_convergence or dres_convergence
    from a single ensemble. Basically, the different slices the algorithm
    needs are generated here.

    Parameters
    ----------

    ensemble : :class:`~MDAnalysis.core.universe.Universe` object
        Input ensemble

    window_size : int
        size of the window (in number of frames) to be used

    select : str
        Atom selection string in the MDAnalysis format. Default is "name CA"

    Returns
    -------

    tmp_ensembles :
        The original ensemble is divided into different ensembles, each being
        a window_size-long slice of the original ensemble. The last
        ensemble will be bigger if the length of the input ensemble
        is not exactly divisible by window_size.

    """

    ens_size = ensemble.trajectory.timeseries(
        ensemble.select_atoms(select), order="fac"
    ).shape[0]

    rest_slices = ens_size // window_size
    residuals = ens_size % window_size
    slices_n = [0]

    tmp_ensembles = []

    for rs in range(rest_slices - 1):
        slices_n.append(slices_n[-1] + window_size)
    slices_n.append(slices_n[-1] + residuals + window_size)

    for s, sl in enumerate(slices_n[:-1]):
        tmp_ensembles.append(
            mda.Universe(
                ensemble.filename,
                ensemble.trajectory.timeseries(order="fac")[
                    slices_n[s] : slices_n[s + 1], :, :
                ],
                format=MemoryReader,
            )
        )

    return tmp_ensembles


def hes(
    ensembles,
    select="name CA",
    cov_estimator="shrinkage",
    weights="mass",
    align=False,
    estimate_error=False,
    bootstrapping_samples=100,
    calc_diagonal=False,
):
    r"""Calculates the Harmonic Ensemble Similarity (HES) between ensembles.

    The HES is calculated with the symmetrized version of Kullback-Leibler
    divergence as described in :footcite:p:`Tiberti2015`.

    Parameters
    ----------
    ensembles : list
        List of Universe objects for similarity measurements.
    select : str, optional
        Atom selection string in the MDAnalysis format. Default is "name CA"
    cov_estimator : str, optional
        Covariance matrix estimator method, either shrinkage, `shrinkage`,
        or Maximum Likelyhood, `ml`. Default is shrinkage.
    weights : str/array_like, optional
        specify optional weights. If ``mass`` then chose masses of ensemble atoms
    align : bool, optional
        Whether to align the ensembles before calculating their similarity.
        Note: this changes the ensembles in-place, and will thus leave your
        ensembles in an altered state.
        (default is False)
    estimate_error : bool, optional
        Whether to perform error estimation (default is False).
    bootstrapping_samples : int, optional
        Number of times the similarity matrix will be bootstrapped (default
        is 100), only if estimate_error is True.
    calc_diagonal : bool, optional
        Whether to calculate the diagonal of the similarity scores
        (i.e. the similarities of every ensemble against itself).
        If this is False (default), 0.0 will be used instead.

    Returns
    -------
    hes, details : numpy.array, dictionary
        Harmonic similarity measurements between each pair of ensembles,
        and dict containing mean and covariance matrix for each ensemble

    Notes
    -----
    The method assumes that each ensemble is derived from a multivariate normal
    distribution. The mean and covariance matrix are, thus, estimatated from
    the distribution of each ensemble and used for comparision by the
    symmetrized version of Kullback-Leibler divergence defined as:

    .. math::
       D_{KL}(P(x) || Q(x)) =
           \int_{-\infty}^{\infty}P(x_i) ln(P(x_i)/Q(x_i)) =
           \langle{}ln(P(x))\rangle{}_P - \langle{}ln(Q(x))\rangle{}_P


    where the :math:`\langle{}.\rangle{}_P` denotes an expectation
    calculated under the distribution :math:`P`.

    For each ensemble, the  mean conformation is estimated as the average over
    the ensemble, and the covariance matrix is calculated by default using a
    shrinkage estimation method (or by a maximum-likelihood method,
    optionally).

    Note that the symmetrized version of the Kullback-Leibler divergence has no
    upper bound (unlike the Jensen-Shannon divergence used by for instance CES and DRES).

    When using this similarity measure, consider whether you want to align
    the ensembles first (see example below).

    Example
    -------

    To calculate the Harmonic Ensemble similarity, two ensembles are created
    as Universe objects from a topology file and two trajectories. The
    topology- and trajectory files used are obtained from the MDAnalysis
    test suite for two different simulations of the protein AdK.
    You can use the ``align=True`` option to align the ensembles first. This will
    align everything to the current timestep in the first ensemble. Note that
    this changes the ``ens1`` and ``ens2`` objects:

        >>> from MDAnalysis import Universe
        >>> import MDAnalysis.analysis.encore as encore
        >>> from MDAnalysis.tests.datafiles import PSF, DCD, DCD2
        >>> ens1 = Universe(PSF, DCD)
        >>> ens2 = Universe(PSF, DCD2)
        >>> HES, details = encore.hes([ens1, ens2])
        >>> print(HES)
        [[       0.         38279540.04524205]
         [38279540.04524205        0.        ]]
        >>> print(encore.hes([ens1, ens2], align=True)[0])
        [[   0.         6889.89729056]
         [6889.89729056    0.        ]]

    Alternatively, for greater flexibility in how the alignment should be done
    you can call use an :class:`~MDAnalysis.analysis.align.AlignTraj` object
    manually:

        >>> from MDAnalysis import Universe
        >>> import MDAnalysis.analysis.encore as encore
        >>> from MDAnalysis.tests.datafiles import PSF, DCD, DCD2
        >>> from MDAnalysis.analysis import align
        >>> ens1 = Universe(PSF, DCD)
        >>> ens2 = Universe(PSF, DCD2)
        >>> _ = align.AlignTraj(ens1, ens1, select="name CA", in_memory=True).run()
        >>> _ = align.AlignTraj(ens2, ens1, select="name CA", in_memory=True).run()
        >>> print(encore.hes([ens1, ens2])[0])
        [[   0.         6889.89729056]
         [6889.89729056    0.        ]]


    .. versionchanged:: 1.0.0
       ``hes`` doesn't accept the `details` argument anymore, it always returns
       the details of the calculation instead, in the form of a dictionary

    """

    if (
        not isinstance(weights, (list, tuple, np.ndarray))
        and weights == "mass"
    ):
        weights = ["mass" for _ in range(len(ensembles))]
    elif weights is not None:
        if len(weights) != len(ensembles):
            raise ValueError("need weights for every ensemble")
    else:
        weights = [None for _ in range(len(ensembles))]

    # Ensure in-memory trajectories either by calling align
    # with in_memory=True or by directly calling transfer_to_memory
    # on the universe.
    if align:
        for e, w in zip(ensembles, weights):
            mda.analysis.align.AlignTraj(
                e, ensembles[0], select=select, weights=w, in_memory=True
            ).run()
    else:
        for ensemble in ensembles:
            ensemble.transfer_to_memory()

    if calc_diagonal:
        pairs_indices = list(trm_indices_diag(len(ensembles)))
    else:
        pairs_indices = list(trm_indices_nodiag(len(ensembles)))

    logging.info("Chosen metric: Harmonic similarity")
    if cov_estimator == "shrinkage":
        covariance_estimator = shrinkage_covariance_estimator
        logging.info("    Covariance matrix estimator: Shrinkage")
    elif cov_estimator == "ml":
        covariance_estimator = ml_covariance_estimator
        logging.info("    Covariance matrix estimator: Maximum Likelihood")
    else:
        logging.error(
            "Covariance estimator {0} is not supported. "
            "Choose between 'shrinkage' and 'ml'.".format(cov_estimator)
        )
        return None

    out_matrix_eln = len(ensembles)

    xs = []
    sigmas = []

    if estimate_error:
        data = []
        ensembles_list = []
        for i, ensemble in enumerate(ensembles):
            ensembles_list.append(
                get_ensemble_bootstrap_samples(
                    ensemble, samples=bootstrapping_samples
                )
            )
        for t in range(bootstrapping_samples):
            logging.info("The coordinates will be bootstrapped.")

            xs = []
            sigmas = []
            values = np.zeros((out_matrix_eln, out_matrix_eln))
            for i, e_orig in enumerate(ensembles):
                xs.append(
                    np.average(
                        ensembles_list[i][t].trajectory.timeseries(
                            e_orig.select_atoms(select), order=("fac")
                        ),
                        axis=0,
                    ).flatten()
                )
                sigmas.append(
                    covariance_matrix(
                        ensembles_list[i][t],
                        weights=weights[i],
                        estimator=covariance_estimator,
                        select=select,
                    )
                )

            for pair in pairs_indices:
                value = harmonic_ensemble_similarity(
                    x1=xs[pair[0]],
                    x2=xs[pair[1]],
                    sigma1=sigmas[pair[0]],
                    sigma2=sigmas[pair[1]],
                )
                values[pair[0], pair[1]] = value
                values[pair[1], pair[0]] = value
            data.append(values)
        avgs = np.average(data, axis=0)
        stds = np.std(data, axis=0)

        return (avgs, stds)

    # Calculate the parameters for the multivariate normal distribution
    # of each ensemble
    values = np.zeros((out_matrix_eln, out_matrix_eln))

    for e, w in zip(ensembles, weights):
        # Extract coordinates from each ensemble
        coordinates_system = e.trajectory.timeseries(
            e.select_atoms(select), order="fac"
        )

        # Average coordinates in each system
        xs.append(np.average(coordinates_system, axis=0).flatten())

        # Covariance matrices in each system
        sigmas.append(
            covariance_matrix(
                e, weights=w, estimator=covariance_estimator, select=select
            )
        )

    for i, j in pairs_indices:
        value = harmonic_ensemble_similarity(
            x1=xs[i], x2=xs[j], sigma1=sigmas[i], sigma2=sigmas[j]
        )
        values[i, j] = value
        values[j, i] = value

    # Save details as required
    details = {}
    for i in range(out_matrix_eln):
        details["ensemble{0:d}_mean".format(i + 1)] = xs[i]
        details["ensemble{0:d}_covariance_matrix".format(i + 1)] = sigmas[i]

    return values, details


def ces(
    ensembles,
    select="name CA",
    clustering_method=None,
    distance_matrix=None,
    estimate_error=False,
    bootstrapping_samples=10,
    ncores=1,
    calc_diagonal=False,
    allow_collapsed_result=True,
):
    """

    Calculates the Clustering Ensemble Similarity (CES) between ensembles
    using the Jensen-Shannon divergence as described in
    :footcite:p:`Tiberti2015`.

    Parameters
    ----------

    ensembles : list
        List of ensemble objects for similarity measurements

    select : str, optional
        Atom selection string in the MDAnalysis format. Default is "name CA"

    clustering_method :
        A single or a list of instances of the
        :class:`MDAnalysis.analysis.encore.clustering.ClusteringMethod` classes
        from the clustering module. Different parameters for the same clustering
        method can be explored by adding different instances of the same
        clustering class. Clustering methods options are the
        Affinity Propagation (default), the DBSCAN and the KMeans. The latter
        two methods need the sklearn python module installed.

    distance_matrix : encore.utils.TriangularMatrix
        Distance matrix clustering methods. If this parameter
        is not supplied the matrix will be calculated on the fly.

    estimate_error :  bool, optional
        Whether to perform error estimation (default is False).
        Only bootstrapping mode is supported.

    bootstrapping_samples : int, optional
        number of samples to be used for estimating error.

    ncores : int, optional
        Maximum number of cores to be used (default is 1).

    calc_diagonal : bool, optional
        Whether to calculate the diagonal of the similarity scores
        (i.e. the similarities of every ensemble against itself).
        If this is False (default), 0.0 will be used instead.

    allow_collapsed_result: bool, optional
        Whether a return value of a list of one value should be collapsed
        into just the value.



    Returns
    -------

    ces, details : numpy.array, numpy.array

        ces contains the similarity values, arranged in a numpy.array.
        If only one clustering_method is provided the output will be a
        2-dimensional square symmetrical numpy.array. The order of the matrix
        elements depends on the order of the input ensembles: for instance, if

            ensemble = [ens1, ens2, ens3]

        the matrix elements [0,2] and [2,0] will both contain the similarity
        value between ensembles ens1 and ens3.
        Elaborating on the previous example, if *n* ensembles are given and *m*
        clustering_methods are provided the output will be a list of *m* arrays
        ordered by the input sequence of methods, each with a *n*x*n*
        symmetrical similarity matrix.

        details contains information on the clustering: the individual size of
        each cluster, the centroids and the frames associated with each cluster.


    Notes
    -----

    In the Jensen-Shannon divergence the upper bound of ln(2) signifies
    no similarity between the two ensembles, the lower bound, 0.0,
    signifies identical ensembles.

    To calculate the CES, the affinity propagation method (or others, if
    specified) is used to partition the whole space of conformations. The
    population of each ensemble in each cluster is then taken as a probability
    density function. Different probability density functions from each
    ensemble are finally compared using the Jensen-Shannon divergence measure.

    Examples
    --------
    To calculate the Clustering Ensemble similarity, two ensembles are
    created as Universe object using a topology file and two trajectories. The
    topology- and trajectory files used are obtained from the MDAnalysis
    test suite for two different simulations of the protein AdK.
    To use a different clustering method, set the parameter clustering_method
    (Note that the sklearn module must be installed). Likewise, different parameters
    for the same clustering method can be explored by adding different
    instances of the same clustering class.
    Here the simplest case of just two instances of :class:`Universe` is illustrated:

        >>> from MDAnalysis import Universe
        >>> import MDAnalysis.analysis.encore as encore
        >>> from MDAnalysis.tests.datafiles import PSF, DCD, DCD2
        >>> ens1 = Universe(PSF, DCD)
        >>> ens2 = Universe(PSF, DCD2)
        >>> CES, details = encore.ces([ens1,ens2])
        >>> print(CES)
        [[0.         0.68070702]
         [0.68070702 0.        ]]
        >>> CES, details = encore.ces([ens1,ens2],
        ...                           clustering_method = [encore.DBSCAN(eps=0.45),
        ...                                                encore.DBSCAN(eps=0.50)])
        >>> print("eps=0.45: ", CES[0])
        eps=0.45:  [[0.         0.20447236]
         [0.20447236 0.        ]]

        >>> print("eps=0.5: ", CES[1])
        eps=0.5:  [[0.         0.25331629]
         [0.25331629 0.        ]]

    """
    if clustering_method is None:
        clustering_method = AffinityPropagationNative(
            preference=-1.0,
            max_iter=500,
            convergence_iter=50,
            damping=0.9,
            add_noise=True,
        )

    for ensemble in ensembles:
        ensemble.transfer_to_memory()

    if calc_diagonal:
        pairs_indices = list(trm_indices_diag(len(ensembles)))
    else:
        pairs_indices = list(trm_indices_nodiag(len(ensembles)))

    clustering_methods = clustering_method
    if not hasattr(clustering_method, "__iter__"):
        clustering_methods = [clustering_method]

    any_method_accept_distance_matrix = np.any(
        [method.accepts_distance_matrix for method in clustering_methods]
    )
    all_methods_accept_distance_matrix = np.all(
        [method.accepts_distance_matrix for method in clustering_methods]
    )

    # Register which ensembles the samples belong to
    ensemble_assignment = []
    for i, ensemble in enumerate(ensembles):
        ensemble_assignment += [i + 1] * len(ensemble.trajectory)

    # Calculate distance matrix if not provided
    if any_method_accept_distance_matrix and not distance_matrix:
        distance_matrix = get_distance_matrix(
            merge_universes(ensembles), select=select, ncores=ncores
        )
    if estimate_error:
        if any_method_accept_distance_matrix:
            distance_matrix = get_distance_matrix_bootstrap_samples(
                distance_matrix,
                ensemble_assignment,
                samples=bootstrapping_samples,
                ncores=ncores,
            )
        if not all_methods_accept_distance_matrix:
            ensembles_list = []
            for i, ensemble in enumerate(ensembles):
                ensembles_list.append(
                    get_ensemble_bootstrap_samples(
                        ensemble, samples=bootstrapping_samples
                    )
                )
            ensembles = []
            for j in range(bootstrapping_samples):
                ensembles.append([])
                for i, e in enumerate(ensembles_list):
                    ensembles[-1].append(e[j])
        else:
            # if all methods accept distances matrices, duplicate
            # ensemble so that it matches size of distance matrices
            # (no need to resample them since they will not be used)
            ensembles = [ensembles] * bootstrapping_samples

    # Call clustering procedure
    ccs = cluster(
        ensembles,
        method=clustering_methods,
        select=select,
        distance_matrix=distance_matrix,
        ncores=ncores,
        allow_collapsed_result=False,
    )

    # Do error analysis
    if estimate_error:
        k = 0
        values = {}
        avgs = []
        stds = []
        for i, p in enumerate(clustering_methods):
            failed_runs = 0
            values[i] = []
            for j in range(bootstrapping_samples):
                if ccs[k].clusters is None:
                    failed_runs += 1
                    k += 1
                    continue
                values[i].append(
                    np.zeros((len(ensembles[j]), len(ensembles[j])))
                )

                for pair in pairs_indices:
                    # Calculate dJS
                    this_djs = clustering_ensemble_similarity(
                        ccs[k],
                        ensembles[j][pair[0]],
                        pair[0] + 1,
                        ensembles[j][pair[1]],
                        pair[1] + 1,
                        select=select,
                    )
                    values[i][-1][pair[0], pair[1]] = this_djs
                    values[i][-1][pair[1], pair[0]] = this_djs
                k += 1
            outs = np.array(values[i])
            avgs.append(np.average(outs, axis=0))
            stds.append(np.std(outs, axis=0))

        if hasattr(clustering_method, "__iter__"):
            pass
        else:
            avgs = avgs[0]
            stds = stds[0]

        return avgs, stds

    values = []
    details = {}
    for i, p in enumerate(clustering_methods):
        if ccs[i].clusters is None:
            continue
        else:
            values.append(np.zeros((len(ensembles), len(ensembles))))

            for pair in pairs_indices:
                # Calculate dJS
                this_val = clustering_ensemble_similarity(
                    ccs[i],
                    ensembles[pair[0]],
                    pair[0] + 1,
                    ensembles[pair[1]],
                    pair[1] + 1,
                    select=select,
                )
                values[-1][pair[0], pair[1]] = this_val
                values[-1][pair[1], pair[0]] = this_val

    details["clustering"] = ccs

    if allow_collapsed_result and not hasattr(clustering_method, "__iter__"):
        values = values[0]

    return values, details


def dres(
    ensembles,
    select="name CA",
    dimensionality_reduction_method=None,
    distance_matrix=None,
    nsamples=1000,
    estimate_error=False,
    bootstrapping_samples=100,
    ncores=1,
    calc_diagonal=False,
    allow_collapsed_result=True,
):
    """

    Calculates the Dimensional Reduction Ensemble Similarity (DRES) between
    ensembles using the Jensen-Shannon divergence as described in
    :footcite:p:`Tiberti2015`.


    Parameters
    ----------

    ensembles : list
        List of ensemble objects for similarity measurements

    select : str, optional
        Atom selection string in the MDAnalysis format. Default is "name CA"

    dimensionality_reduction_method :
        A single or a list of instances of the DimensionalityReductionMethod
        classes from the dimensionality_reduction module. Different parameters
        for the same method can be explored by adding different instances of
        the same dimensionality reduction class. Provided methods are the
        Stochastic Proximity Embedding (default) and the Principal Component
        Analysis.

    distance_matrix : encore.utils.TriangularMatrix
        conformational distance matrix, It will be calculated on the fly
        from the ensemble data if it is not provided.

    nsamples : int, optional
        Number of samples to be drawn from the ensembles (default is 1000).
        This is used to resample the density estimates and calculate the
        Jensen-Shannon divergence between ensembles.

    estimate_error : bool, optional
        Whether to perform error estimation (default is False)

    bootstrapping_samples : int, optional
        number of samples to be used for estimating error.

    ncores : int, optional
        Maximum number of cores to be used (default is 1).

    calc_diagonal : bool, optional
        Whether to calculate the diagonal of the similarity scores
        (i.e. the simlarities of every ensemble against itself).
        If this is False (default), 0.0 will be used instead.

    allow_collapsed_result: bool, optional
        Whether a return value of a list of one value should be collapsed
        into just the value.


    Returns
    -------

    dres, details : numpy.array, numpy.array
        dres contains the similarity values, arranged in numpy.array.
        If one number of dimensions is provided as an integer,
        the output will be a 2-dimensional square symmetrical numpy.array.
        The order of the matrix elements depends on the order of the
        input ensemble: for instance, if

            ensemble = [ens1, ens2, ens3]

        then the matrix elements [0,2] and [2,0] will both contain the
        similarity value between ensembles ens1 and ens3.
        Elaborating on the previous example, if *n* ensembles are given and *m*
        methods are provided the output will be a list of *m* arrays
        ordered by the input sequence of methods, each with a *n*x*n*
        symmetrical similarity matrix.

        details provide an array of the reduced_coordinates.

    Notes
    -----

    To calculate the similarity, the method first projects the ensembles into
    lower dimensions by using the Stochastic Proximity Embedding (or others)
    algorithm. A gaussian kernel-based density estimation method is then used
    to estimate the probability density for each ensemble which is then used
    to compute the Jensen-Shannon divergence between each pair of ensembles.

    In the Jensen-Shannon divergence the upper bound of ln(2) signifies
    no similarity between the two ensembles, the lower bound, 0.0,
    signifies identical ensembles. However, due to the stochastic nature of
    the dimensional reduction in :func:`dres`, two identical ensembles will
    not necessarily result in an exact 0.0 estimate of the similarity but
    will be very close. For the same reason, calculating the similarity with
    the :func:`dres` twice will not result in two identical numbers; small
    differences have to be expected.

    Examples
    --------

    To calculate the Dimensional Reduction Ensemble similarity, two ensembles
    are created as Universe objects from a topology file and two trajectories.
    The topology- and trajectory files used are obtained from the MDAnalysis
    test suite for two different simulations of the protein AdK.
    To use a different dimensional reduction methods, simply set the
    parameter dimensionality_reduction_method. Likewise, different parameters
    for the same clustering method can be explored by adding different
    instances of the same method  class.
    Here the simplest case of comparing just two instances of :class:`Universe` is
    illustrated:

        >>> from MDAnalysis import Universe
        >>> import MDAnalysis.analysis.encore as encore
        >>> from MDAnalysis.tests.datafiles import PSF, DCD, DCD2
        >>> ens1 = Universe(PSF,DCD)
        >>> ens2 = Universe(PSF,DCD2)
        >>> DRES, details = encore.dres([ens1,ens2])
        >>> PCA_method = encore.PrincipalComponentAnalysis(dimension=2)
        >>> DRES, details = encore.dres([ens1,ens2],
        ...                             dimensionality_reduction_method=PCA_method)

    In addition to the quantitative similarity estimate, the dimensional
    reduction can easily be visualized, see the ``Example`` section in
    :mod:`MDAnalysis.analysis.encore.dimensionality_reduction.reduce_dimensionality``

    """
    if dimensionality_reduction_method is None:
        dimensionality_reduction_method = StochasticProximityEmbeddingNative(
            dimension=3,
            distance_cutoff=1.5,
            min_lam=0.1,
            max_lam=2.0,
            ncycle=100,
            nstep=10000,
        )

    for ensemble in ensembles:
        ensemble.transfer_to_memory()

    if calc_diagonal:
        pairs_indices = list(trm_indices_diag(len(ensembles)))
    else:
        pairs_indices = list(trm_indices_nodiag(len(ensembles)))

    dimensionality_reduction_methods = dimensionality_reduction_method
    if not hasattr(dimensionality_reduction_method, "__iter__"):
        dimensionality_reduction_methods = [dimensionality_reduction_method]

    any_method_accept_distance_matrix = np.any(
        [
            method.accepts_distance_matrix
            for method in dimensionality_reduction_methods
        ]
    )
    all_methods_accept_distance_matrix = np.all(
        [
            method.accepts_distance_matrix
            for method in dimensionality_reduction_methods
        ]
    )

    # Register which ensembles the samples belong to
    ensemble_assignment = []
    for i, ensemble in enumerate(ensembles):
        ensemble_assignment += [i + 1] * len(ensemble.trajectory)

    # Calculate distance matrix if not provided
    if any_method_accept_distance_matrix and not distance_matrix:
        distance_matrix = get_distance_matrix(
            merge_universes(ensembles), select=select, ncores=ncores
        )
    if estimate_error:
        if any_method_accept_distance_matrix:
            distance_matrix = get_distance_matrix_bootstrap_samples(
                distance_matrix,
                ensemble_assignment,
                samples=bootstrapping_samples,
                ncores=ncores,
            )
        if not all_methods_accept_distance_matrix:
            ensembles_list = []
            for i, ensemble in enumerate(ensembles):
                ensembles_list.append(
                    get_ensemble_bootstrap_samples(
                        ensemble, samples=bootstrapping_samples
                    )
                )
            ensembles = []
            for j in range(bootstrapping_samples):
                ensembles.append(
                    ensembles_list[i, j]
                    for i in range(ensembles_list.shape[0])
                )
        else:
            # if all methods accept distances matrices, duplicate
            # ensemble so that it matches size of distance matrices
            # (no need to resample them since they will not be used)
            ensembles = [ensembles] * bootstrapping_samples

    # Call dimensionality reduction procedure
    coordinates, dim_red_details = reduce_dimensionality(
        ensembles,
        method=dimensionality_reduction_methods,
        select=select,
        distance_matrix=distance_matrix,
        ncores=ncores,
        allow_collapsed_result=False,
    )

    details = {}
    details["reduced_coordinates"] = coordinates
    details["dimensionality_reduction_details"] = dim_red_details

    if estimate_error:
        k = 0
        values = {}
        avgs = []
        stds = []
        for i, method in enumerate(dimensionality_reduction_methods):
            values[i] = []
            for j in range(bootstrapping_samples):

                values[i].append(
                    np.zeros((len(ensembles[j]), len(ensembles[j])))
                )

                kdes, resamples, embedded_ensembles = gen_kde_pdfs(
                    coordinates[k],
                    ensemble_assignment,
                    len(ensembles[j]),
                    nsamples=nsamples,
                )

                for pair in pairs_indices:
                    this_value = dimred_ensemble_similarity(
                        kdes[pair[0]],
                        resamples[pair[0]],
                        kdes[pair[1]],
                        resamples[pair[1]],
                    )
                    values[i][-1][pair[0], pair[1]] = this_value
                    values[i][-1][pair[1], pair[0]] = this_value

                k += 1
            outs = np.array(values[i])
            avgs.append(np.average(outs, axis=0))
            stds.append(np.std(outs, axis=0))

        if hasattr(dimensionality_reduction_method, "__iter__"):
            pass
        else:
            avgs = avgs[0]
            stds = stds[0]

        return avgs, stds

    values = []

    for i, method in enumerate(dimensionality_reduction_methods):

        values.append(np.zeros((len(ensembles), len(ensembles))))
        kdes, resamples, embedded_ensembles = gen_kde_pdfs(
            coordinates[i],
            ensemble_assignment,
            len(ensembles),
            nsamples=nsamples,
        )

        for pair in pairs_indices:
            this_value = dimred_ensemble_similarity(
                kdes[pair[0]],
                resamples[pair[0]],
                kdes[pair[1]],
                resamples[pair[1]],
            )
            values[-1][pair[0], pair[1]] = this_value
            values[-1][pair[1], pair[0]] = this_value

    if allow_collapsed_result and not hasattr(
        dimensionality_reduction_method, "__iter__"
    ):
        values = values[0]

    return values, details


def ces_convergence(
    original_ensemble,
    window_size,
    select="name CA",
    clustering_method=None,
    ncores=1,
):
    """
    Use the CES to evaluate the convergence of the ensemble/trajectory.
    CES will be calculated between the whole trajectory contained in an
    ensemble and windows of such trajectory of increasing sizes, so that
    the similarity values should gradually drop to zero. The rate at which
    the value reach zero will be indicative of how much the trajectory
    keeps on resampling the same regions of the conformational space, and
    therefore of convergence.


    Parameters
    ----------

    original_ensemble : :class:`~MDAnalysis.core.universe.Universe` object
        ensemble containing the trajectory whose convergence has to estimated

    window_size : int
        Size of window to be used, in number of frames

    select : str, optional
        Atom selection string in the MDAnalysis format. Default is "name CA"

    clustering_method : MDAnalysis.analysis.encore.clustering.ClusteringMethod
        A single or a list of instances of the ClusteringMethod classes from
        the clustering module. Different parameters for the same clustering
        method can be explored by adding different instances of the same
        clustering class.

    ncores : int, optional
        Maximum number of cores to be used (default is 1).


    Returns
    -------

    out : np.array
        array of shape (number_of_frames / window_size, preference_values).


    Example
    --------
    To calculate the convergence of a trajectory using the clustering ensemble
    similarity method a Universe object is created from a topology file and the
    trajectory. The topology- and trajectory files used are obtained from the
    MDAnalysis test suite for two different simulations of the protein AdK.
    Here the simplest case of evaluating the convergence is illustrated by
    splitting the trajectory into a window_size of 10 frames:

        >>> from MDAnalysis import Universe
        >>> import MDAnalysis.analysis.encore as encore
        >>> from MDAnalysis.tests.datafiles import PSF, DCD, DCD2
        >>> ens1 = Universe(PSF,DCD)
        >>> ces_conv = encore.ces_convergence(ens1, 10)
        >>> print(ces_conv)
        [[0.48194205]
         [0.40284672]
         [0.31699026]
         [0.25220447]
         [0.19829817]
         [0.14642725]
         [0.09911411]
         [0.05667391]
         [0.        ]]

    """
    if clustering_method is None:
        clustering_method = AffinityPropagationNative(
            preference=-1.0,
            max_iter=500,
            convergence_iter=50,
            damping=0.9,
            add_noise=True,
        )

    ensembles = prepare_ensembles_for_convergence_increasing_window(
        original_ensemble, window_size, select=select
    )

    ccs = cluster(
        ensembles,
        select=select,
        method=clustering_method,
        allow_collapsed_result=False,
        ncores=ncores,
    )

    out = []
    for cc in ccs:
        if cc.clusters is None:
            continue
        out.append(np.zeros(len(ensembles)))
        for j, ensemble in enumerate(ensembles):
            out[-1][j] = cumulative_clustering_ensemble_similarity(
                cc, len(ensembles), j + 1
            )

    out = np.array(out).T
    return out


def dres_convergence(
    original_ensemble,
    window_size,
    select="name CA",
    dimensionality_reduction_method=None,
    nsamples=1000,
    ncores=1,
):
    """
    Use the DRES to evaluate the convergence of the ensemble/trajectory.
    DRES will be calculated between the whole trajectory contained in an
    ensemble and windows of such trajectory of increasing sizes, so that
    the similarity values should gradually drop to zero. The rate at which
    the value reach zero will be indicative of how much the trajectory
    keeps on resampling the same ares of the conformational space, and
    therefore of convergence.

    Parameters
    ----------

    original_ensemble : :class:`~MDAnalysis.core.universe.Universe` object
        ensemble containing the trajectory whose convergence has to estimated

    window_size : int
        Size of window to be used, in number of frames

    select : str, optional
        Atom selection string in the MDAnalysis format. Default is "name CA"

    dimensionality_reduction_method :
        A single or a list of instances of the DimensionalityReductionMethod
        classes from the dimensionality_reduction module. Different parameters
        for the same method can be explored by adding different instances of
        the same dimensionality reduction class.

    nsamples : int, optional
        Number of samples to be drawn from the ensembles (default is 1000).
        This is akin to the nsamples parameter of dres().

    ncores  : int, optional
        Maximum number of cores to be used (default is 1).


    Returns
    -------

    out : np.array
        array of shape (number_of_frames / window_size, preference_values).



    Example
    --------
    To calculate the convergence of a trajectory using the DRES
    method, a Universe object is created from a topology file and the
    trajectory. The topology- and trajectory files used are obtained from the
    MDAnalysis test suite for two different simulations of the protein AdK.
    Here the simplest case of evaluating the convergence is illustrated by
    splitting the trajectory into a window_size of 10 frames:

        >>> from MDAnalysis import Universe
        >>> import MDAnalysis.analysis.encore as encore
        >>> from MDAnalysis.tests.datafiles import PSF, DCD, DCD2
        >>> ens1 = Universe(PSF,DCD)
        >>> dres_conv = encore.dres_convergence(ens1, 10)

    Here, the rate at which the values reach zero will be indicative of how
    much the trajectory keeps on resampling the same ares of the conformational
    space, and therefore of convergence.
    """
    if dimensionality_reduction_method is None:
        dimensionality_reduction_method = StochasticProximityEmbeddingNative(
            dimension=3,
            distance_cutoff=1.5,
            min_lam=0.1,
            max_lam=2.0,
            ncycle=100,
            nstep=10000,
        )

    ensembles = prepare_ensembles_for_convergence_increasing_window(
        original_ensemble, window_size, select=select
    )

    coordinates, dimred_details = reduce_dimensionality(
        ensembles,
        select=select,
        method=dimensionality_reduction_method,
        allow_collapsed_result=False,
        ncores=ncores,
    )

    ensemble_assignment = []
    for i, ensemble in enumerate(ensembles):
        ensemble_assignment += [i + 1] * len(ensemble.trajectory)
    ensemble_assignment = np.array(ensemble_assignment)

    out = []
    for i, _ in enumerate(coordinates):

        out.append(np.zeros(len(ensembles)))

        kdes, resamples, embedded_ensembles = cumulative_gen_kde_pdfs(
            coordinates[i],
            ensemble_assignment=ensemble_assignment,
            nensembles=len(ensembles),
            nsamples=nsamples,
        )

        for j, ensemble in enumerate(ensembles):
            out[-1][j] = dimred_ensemble_similarity(
                kdes[-1], resamples[-1], kdes[j], resamples[j]
            )

    out = np.array(out).T
    return out



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

__all__ = [
    "align",
    "backends",
    "base",
    "contacts",
    "density",
    "distances",
    "diffusionmap",
    "dihedrals",
    "distances",
    "dielectric",
    "gnm",
    "hbonds",
    "helix_analysis",
    "hole2",
    "hydrogenbonds",
    "leaflet",
    "lineardensity",
    "msd",
    "nuclinfo",
    "nucleicacids",
    "polymer",
    "pca",
    "psa",
    "rdf",
    "results",
    "rms",
    "waterdynamics",
]



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

"""\
Core Topology object --- :mod:`MDAnalysis.core.topology`
========================================================

.. versionadded:: 0.16.0

:class:`Topology` is the core object that holds all topology information.

TODO: Add in-depth discussion.

Notes
-----
For developers: In MDAnalysis 0.16.0 this new topology system was
introduced and discussed as issue `#363`_; this issue contains key
information and discussions on the new system. The issue number *363*
is also being used as a short-hand in discussions to refer to the new
topology system.


.. _`#363`: https://github.com/MDAnalysis/mdanalysis/issues/363

Classes
-------

.. autoclass:: Topology
   :members:
.. autoclass:: TransTable
   :members:

Helper functions
----------------

.. autofunction:: make_downshift_arrays

"""
import contextlib

import numpy as np
import typing

from .topologyattrs import Atomindices, Resindices, Segindices
from ..exceptions import NoDataError


def make_downshift_arrays(upshift, nparents):
    """From an upwards translation table, create the opposite direction

    Turns a many to one mapping (eg atoms to residues) to a one to many mapping
    (residues to atoms)

    Parameters
    ----------
    upshift : array_like
        Array of integers describing which parent each item belongs to
    nparents : integer
        Total number of parents that exist.

    Returns
    -------
    downshift : array_like (dtype object)
        An array of arrays, each containing the indices of the children
        of each parent.  Length `nparents` + 1

    Examples
    --------

    To find the residue to atom mappings for a given atom to residue mapping:

    >>> import numpy as np
    >>> import MDAnalysis as mda
    >>> from MDAnalysis.core.topology import make_downshift_arrays
    >>> atom2res = np.array([0, 1, 0, 2, 2, 0, 2])
    >>> make_downshift_arrays(atom2res, 3)
    array([array([0, 2, 5]), array([1]), array([3, 4, 6]), None], dtype=object)

    Entry 0 corresponds to residue 0 and says that this contains atoms 0, 2 & 5

    Notes
    -----
    The final entry in the return array will be ``None`` to ensure that the
    dtype of the array is :class:`object`.

    .. warning:: This means negative indexing should **never**
                 be used with these arrays.
    """
    if not len(upshift):
        return np.array([], dtype=object)

    # mergesort for a stable ordered array for the same value.
    order = np.argsort(upshift, kind="mergesort")

    upshift_sorted = upshift[order]
    u_values, indices = np.unique(upshift_sorted, return_index=True)

    # reset nparents to the larger one between input and heuristic from data
    # This is useful for creating empty Universe where default value is 1.
    nparents = np.max([nparents, u_values.max() + 1])
    residue_indices = np.zeros(nparents, dtype=int)
    missing_resids = np.sort(np.setdiff1d(np.arange(nparents), u_values))
    indices = np.append(indices, upshift_sorted.shape[0])

    residue_indices[u_values] = indices[1:]

    for missing_resid in missing_resids:
        if missing_resid == 0:
            residue_indices[missing_resid] = 0
        else:
            residue_indices[missing_resid] = residue_indices[missing_resid - 1]

    downshift = np.split(order, residue_indices[:-1])
    # Add None to end of array to force it to be of type Object
    # Without this, a rectangular array gets squashed into a single array
    downshift.append(None)
    return np.array(downshift, dtype=object)


class TransTable(object):
    """Membership tables with methods to translate indices across levels.

    There are three levels; Atom, Residue and Segment.  Each Atom **must**
    belong in a Residue, each Residue **must** belong to a Segment.

    When translating upwards, eg finding which Segment a Residue belongs in,
    a single numpy array is returned.  When translating downwards, two options
    are available; a concatenated result (suffix `_1`) or a list for each parent
    object (suffix `_2d`).

    Parameters
    ----------
    n_atoms : int
        number of atoms in topology
    n_residues : int
        number of residues in topology
    n_segments : int
        number of segments in topology
    atom_resindex : 1-D array
        resindex for each atom in the topology; the number of unique values in
        this array must be <= `n_residues`, and the array must be length
        `n_atoms`; giving None defaults to placing all atoms in residue 0
    residue_segindex : 1-D array
        segindex for each residue in the topology; the number of unique values
        in this array must be <= `n_segments`, and the array must be length
        `n_residues`; giving None defaults to placing all residues in segment 0


    Attributes
    ----------
    n_atoms : int
        number of atoms in topology
    n_residues : int
        number of residues in topology
    n_segments : int
        number of segments in topology
    size : tuple
        tuple ``(n_atoms, n_residues, n_segments)`` describing the shape of
        the TransTable

    .. versionchanged:: 2.3.0
        Lazy building RA and SR.
    """

    def __init__(
        self,
        n_atoms,
        n_residues,
        n_segments,  # Size of tables
        atom_resindex=None,
        residue_segindex=None,  # Contents of tables
    ):
        self.n_atoms = n_atoms
        self.n_residues = n_residues
        self.n_segments = n_segments

        # built atom-to-residue mapping, and vice-versa
        if atom_resindex is None:
            self._AR = np.zeros(n_atoms, dtype=np.intp)
        else:
            self._AR = np.asarray(atom_resindex, dtype=np.intp).copy()
            if not len(self._AR) == n_atoms:
                raise ValueError("atom_resindex must be len n_atoms")
        self._RA = None

        # built residue-to-segment mapping, and vice-versa
        if residue_segindex is None:
            self._RS = np.zeros(n_residues, dtype=np.intp)
        else:
            self._RS = np.asarray(residue_segindex, dtype=np.intp).copy()
            if not len(self._RS) == n_residues:
                raise ValueError("residue_segindex must be len n_residues")
        self._SR = None

    def copy(self):
        """Return a deepcopy of this Transtable"""
        return self.__class__(
            self.n_atoms,
            self.n_residues,
            self.n_segments,
            atom_resindex=self._AR,
            residue_segindex=self._RS,
        )

    @property
    def RA(self):
        if self._RA is None:
            self._RA = make_downshift_arrays(self._AR, self.n_residues)
        return self._RA

    @property
    def SR(self):
        if self._SR is None:
            self._SR = make_downshift_arrays(self._RS, self.n_segments)
        return self._SR

    @property
    def size(self):
        """The shape of the table, ``(n_atoms, n_residues, n_segments)``.

        :meta private:
        """
        return (self.n_atoms, self.n_residues, self.n_segments)

    def atoms2residues(self, aix):
        """Get residue indices for each atom.

        Parameters
        ----------
        aix : array
            atom indices

        Returns
        -------
        rix : array
            residue index for each atom

        """
        return self._AR[aix]

    def residues2atoms_1d(self, rix):
        """Get atom indices collectively represented by given residue indices.

        Parameters
        ----------
        rix : array
            residue indices

        Returns
        -------
        aix : array
            indices of atoms present in residues, collectively

        """
        RA = self.RA
        try:
            return np.concatenate(RA[rix])
        except ValueError:  # rix is not iterable or empty
            # don't accidentally return a view!
            return RA[rix].astype(np.intp, copy=True)

    def residues2atoms_2d(self, rix):
        """Get atom indices represented by each residue index.

        Parameters
        ----------
        rix : array
            residue indices

        Returns
        -------
        raix : list
            each element corresponds to a residue index, in order given in
            `rix`, with each element being an array of the atom indices present
            in that residue

        """
        RA = self.RA
        try:
            return [RA[r].copy() for r in rix]
        except TypeError:
            return [RA[rix].copy()]  # why would this be singular for 2d?

    def residues2segments(self, rix):
        """Get segment indices for each residue.

        Parameters
        ----------
        rix : array
            residue indices

        Returns
        -------
        six : array
            segment index for each residue

        """
        return self._RS[rix]

    def segments2residues_1d(self, six):
        """Get residue indices collectively represented by given segment indices

        Parameters
        ----------
        six : array
            segment indices

        Returns
        -------
        rix : array
            sorted indices of residues present in segments, collectively

        """
        SR = self.SR
        try:
            return np.concatenate(SR[six])
        except ValueError:  # six is not iterable or empty
            # don't accidentally return a view!
            return SR[six].astype(np.intp, copy=True)

    def segments2residues_2d(self, six):
        """Get residue indices represented by each segment index.

        Parameters
        ----------
        six : array
            residue indices

        Returns
        -------
        srix : list
            each element corresponds to a segment index, in order given in
            `six`, with each element being an array of the residue indices
            present in that segment

        """
        SR = self.SR
        try:
            return [SR[s].copy() for s in six]
        except TypeError:
            return [SR[six].copy()]

    # Compound moves, does 2 translations
    def atoms2segments(self, aix):
        """Get segment indices for each atom.

        Parameters
        ----------
        aix : array
            atom indices

        Returns
        -------
        rix : array
            segment index for each atom

        """
        rix = self.atoms2residues(aix)
        return self.residues2segments(rix)

    def segments2atoms_1d(self, six):
        """Get atom indices collectively represented by given segment indices.

        Parameters
        ----------
        six : array
            segment indices

        Returns
        -------
        aix : array
            sorted indices of atoms present in segments, collectively

        """
        rix = self.segments2residues_1d(six)
        return self.residues2atoms_1d(rix)

    def segments2atoms_2d(self, six):
        """Get atom indices represented by each segment index.

        Parameters
        ----------
        six : array
            residue indices

        Returns
        -------
        saix : list
            each element corresponds to a segment index, in order given in
            `six`, with each element being an array of the atom indices present
            in that segment

        """
        # residues in EACH
        rixs = self.segments2residues_2d(six)
        return [self.residues2atoms_1d(rix) for rix in rixs]

    # Move between different groups.
    def move_atom(self, aix, rix):
        """Move aix to be in rix"""
        self._AR[aix] = rix
        self._RA = None

    def move_residue(self, rix, six):
        """Move rix to be in six"""
        self._RS[rix] = six
        self._SR = None

    def add_Residue(self, segidx):
        # segidx - index of parent
        self.n_residues += 1
        self._RA = None
        self._RS = np.concatenate([self._RS, np.array([segidx])])
        self._SR = None

        return self.n_residues - 1

    def add_Segment(self):
        self.n_segments += 1
        self._SR = None
        return self.n_segments - 1

    def __getstate__(self):
        # don't serialize _RA and _SR for performance.
        attrs = self.__dict__
        attrs["_RA"] = None
        attrs["_SR"] = None
        return attrs


class Topology(object):
    """In-memory, array-based topology database.

    The topology model of MDanalysis features atoms, which must each be a
    member of one residue. Each residue, in turn, must be a member of one
    segment. The details of maintaining this heirarchy, and mappings of atoms
    to residues, residues to segments, and vice-versa, are handled internally
    by this object.

    """

    def __init__(
        self,
        n_atoms=1,
        n_res=1,
        n_seg=1,
        attrs=None,
        atom_resindex=None,
        residue_segindex=None,
    ):
        """
        Parameters
        ----------
        n_atoms : int
            number of atoms in topology. Must be larger then 1 at each level
        n_residues : int
            number of residues in topology. Must be larger then 1 at each level
        n_segments : int
            number of segments in topology. Must be larger then 1 at each level
        attrs : TopologyAttr objects
            components of the topology to be included
        atom_resindex : array
            1-D array giving the resindex of each atom in the system
        residue_segindex : array
            1-D array giving the segindex of each residue in the system

        """
        self.tt = TransTable(
            n_atoms,
            n_res,
            n_seg,
            atom_resindex=atom_resindex,
            residue_segindex=residue_segindex,
        )

        if attrs is None:
            attrs = []
        # add core TopologyAttrs that give access to indices
        attrs.extend((Atomindices(), Resindices(), Segindices()))

        # attach the TopologyAttrs
        self.attrs = []
        for topologyattr in attrs:
            self.add_TopologyAttr(topologyattr)

    def copy(self):
        """Return a deepcopy of this Topology"""
        new = self.__class__(1, 1, 1)
        # copy the tt
        new.tt = self.tt.copy()
        # remove indices
        for attr in self.attrs:
            if isinstance(attr, (Atomindices, Resindices, Segindices)):
                continue
            new.add_TopologyAttr(attr.copy())
        return new

    @property
    def n_atoms(self):
        return self.tt.n_atoms

    @property
    def n_residues(self):
        return self.tt.n_residues

    @property
    def n_segments(self):
        return self.tt.n_segments

    def add_TopologyAttr(self, topologyattr):
        """Add a new TopologyAttr to the Topology.

        Parameters
        ----------
        topologyattr : TopologyAttr

        """
        self.attrs.append(topologyattr)
        topologyattr.top = self
        self.__setattr__(topologyattr.attrname, topologyattr)

    def del_TopologyAttr(self, topologyattr):
        """Remove a TopologyAttr from the Topology.

        If it is not present, nothing happens.

        Parameters
        ----------
        topologyattr : TopologyAttr


        .. versionadded:: 2.0.0
        """
        self.__delattr__(topologyattr.attrname)
        self.attrs.remove(topologyattr)

    @property
    def guessed_attributes(self):
        """A list of the guessed attributes in this topology"""
        return filter(
            lambda x: (
                x.is_guessed
                if (not isinstance(x.is_guessed, typing.Container))
                else True in x.is_guessed
            ),
            self.attrs,
        )

    @property
    def read_attributes(self):
        """A list of the attributes read from the topology"""
        return filter(
            lambda x: (
                not x.is_guessed
                if (not isinstance(x.is_guessed, typing.Container))
                else False in x.is_guessed
            ),
            self.attrs,
        )

    def add_Residue(self, segment, **new_attrs):
        """
        Returns
        -------
        residx of the new Residue

        Raises
        ------
        NoDataError
          If not all data was provided.  This error is raised before any changes


        .. versionchanged:: 2.1.0
           Added use of _add_new to TopologyAttr resize
        """
        # Check that all data is here before making any changes
        for attr in self.attrs:
            if not attr.per_object == "residue":
                continue
            if attr.singular not in new_attrs:
                missing = (
                    attr.singular
                    for attr in self.attrs
                    if (
                        attr.per_object == "residue"
                        and attr.singular not in new_attrs
                    )
                )
                raise NoDataError(
                    "Missing the following attributes for the new"
                    " Residue: {}".format(", ".join(missing))
                )

        # Resize topology table
        residx = self.tt.add_Residue(segment.segindex)

        # Add new value to each attribute
        for attr in self.attrs:
            if not attr.per_object == "residue":
                continue
            newval = new_attrs[attr.singular]
            attr._add_new(newval)

        return residx

    def add_Segment(self, **new_attrs):
        """Adds a new Segment to the Topology

        Parameters
        ----------
        new_attrs : dict
          the new attributes for the new segment, eg {'segid': 'B'}

        Raises
        -------
        NoDataError
          if an attribute wasn't specified.

        Returns
        -------
        ix : int
          the idx of the new segment


        .. versionchanged:: 2.1.0
           Added use of _add_new to resize topology attrs
        """
        for attr in self.attrs:
            if attr.per_object == "segment":
                if attr.singular not in new_attrs:
                    missing = (
                        attr.singular
                        for attr in self.attrs
                        if (
                            attr.per_object == "segment"
                            and attr.singular not in new_attrs
                        )
                    )
                    raise NoDataError(
                        "Missing the following attributes for the"
                        " new Segment: {}"
                        "".format(", ".join(missing))
                    )

        segidx = self.tt.add_Segment()

        for attr in self.attrs:
            if not attr.per_object == "segment":
                continue
            newval = new_attrs[attr.singular]
            attr._add_new(newval)

        return segidx



"""
IMDReader --- :mod:`MDAnalysis.coordinates.IMD`
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

This module provides support for reading molecular dynamics simulation data via the 
`Interactive Molecular Dynamics (IMD) protocol v3 <https://imdclient.readthedocs.io/en/latest/protocol_v3.html>`_.
The IMD protocol allows two-way communicating molecular simulation data through a socket.
Via IMD, a simulation engine sends data to a receiver (in this case, the IMDClient) and the receiver can send forces and specific control
requests (such as pausing, resuming, or terminating the simulation) back to the simulation engine. 

.. note::
   This reader only supports IMDv3, which is implemented in GROMACS, LAMMPS, and NAMD at varying
   stages of development. See the `imdclient simulation engine docs`_ for more. 
   While IMDv2 is widely available in simulation engines, it was designed primarily for visualization
   and gaps are allowed in the stream (i.e., an inconsistent number of integrator time steps between transmitted coordinate arrays is allowed)

The :class:`IMDReader` connects to a simulation via a socket and receives coordinate,
velocity, force, and energy data as the simulation progresses. This allows for real-time
monitoring and analysis of ongoing simulations. It uses the `imdclient package <https://github.com/Becksteinlab/imdclient>`_ 
(dependency) to implement the IMDv3 protocol and manage the socket connection and data parsing.

.. seealso::
   :class:`IMDReader`
      Technical details and parameter options for the reader class
   
   `imdclient documentation <https://imdclient.readthedocs.io/>`_
      Complete documentation for the IMDClient package
      
   `IMDClient GitHub repository <https://github.com/Becksteinlab/imdclient>`_
      Source code and development resources

.. _`imdclient simulation engine docs`: https://imdclient.readthedocs.io/en/latest/usage.html

Usage Example
-------------

As an example of reading a stream, after configuring GROMACS to run a simulation with IMDv3 enabled
(see the `imdclient simulation engine docs`_ for 
up-to-date resources on configuring each simulation engine), use the following commands:

.. code-block:: bash

    gmx mdrun -v -nt 4 -imdwait -imdport 8889

The :class:`~MDAnalysis.coordinates.IMD.IMDReader` can then connect to the running simulation and stream data in real time:

.. code-block:: python

    import MDAnalysis as mda
    u = mda.Universe("topol.tpr", "imd://localhost:8889", buffer_size=10*1024*1024)

    print("    time [         position         ] [         velocity         ] [           force          ] [            box           ]")
    sel = u.select_atoms("all")  # Select all atoms; adjust selection as needed
    for ts in u.trajectory:
        print(f'{ts.time:8.3f} {sel[0].position} {sel[0].velocity} {sel[0].force} {u.dimensions[0:3]}')

.. important::
   **Jupyter Notebook Users**: When using IMDReader in Jupyter notebooks, be aware that 
   **kernel restarts will not gracefully close active IMD connections**. This can leave 
   socket connections open, potentially preventing new connections to the same stream.

   Always use ``try/except/finally`` blocks to ensure proper cleanup:

   .. code-block:: python

       import MDAnalysis as mda
       
       try:
           u = mda.Universe("topol.tpr", "imd://localhost:8889")
       except Exception as e:
           print(f"Error during connection: {e}")
       else:
           try:
               # Your analysis code here
               for ts in u.trajectory:
                   # Process each frame
                   pass
           finally:
               # Ensure connection is closed
               u.trajectory.close()

   Always explicitly call ``u.trajectory.close()`` when finished with analysis to 
   ensure connection is closed properly.

Important Limitations
---------------------

.. warning::  
   The IMDReader has some important limitations that are inherent in streaming data.  

Since IMD streams data in real-time from a running simulation, it has fundamental 
constraints that differ from traditional trajectory readers:

* **No random access**: Cannot jump to arbitrary frame numbers or seek backwards  
* **Forward-only access**: You can only move forward through frames as they arrive
* **No trajectory length**: The total number of frames is unknown until the simulation ends
* **Single-use iteration**: Cannot restart or rewind once the stream has been consumed
* **No independent copies**: Cannot create separate reader instances for the same stream
* **No stream restart**: Cannot reconnect or reopen once the connection is closed
* **No bulk operations**: Cannot extract all data at once using timeseries methods
* **Limited multiprocessing**: Cannot split reader across processes for parallel analysis
* **Single client connection**: Only one reader can connect to an IMD stream at a time
* **No trajectory Writing**: Complimentary IMD Writer class is not available for streaming data

.. seealso::
   See :class:`~MDAnalysis.coordinates.base.StreamReaderBase` for technical details.

Multiple Client Connections
---------------------------

The ability to establish multiple simultaneous connections to the same IMD port is 
**MD engine implementation dependent**. Some simulation engines may allow multiple 
clients to connect concurrently, while others may reject or fail additional connection 
attempts.

See the `imdclient simulation engine docs`_ for further details. 

.. important::
   Even when multiple connections are supported by the simulation engine, each connection 
   receives its own independent data stream. These streams may contain different data 
   depending on the simulation engine's configuration, so multiple connections should 
   not be assumed to provide identical data streams.

Classes
-------

.. autoclass:: IMDReader
   :members:
   :inherited-members:

"""

import numpy as np
import logging
import warnings

from MDAnalysis.coordinates import core
from MDAnalysis.lib.util import store_init_arguments
from MDAnalysis.coordinates.base import StreamReaderBase


from packaging.version import Version

MIN_IMDCLIENT_VERSION = Version("0.2.2")

try:
    import imdclient
    from imdclient.IMDClient import IMDClient
    from imdclient.utils import parse_host_port
except ImportError:
    HAS_IMDCLIENT = False
    imdclient_version = Version("0.0.0")

    # Allow building documentation without imdclient
    import types

    class MockIMDClient:
        pass

    imdclient = types.ModuleType("imdclient")
    imdclient.IMDClient = MockIMDClient
    imdclient.__version__ = "0.0.0"

else:
    HAS_IMDCLIENT = True
    imdclient_version = Version(imdclient.__version__)

    # Check for compatibility: currently needs to be >=0.2.2
    if imdclient_version < MIN_IMDCLIENT_VERSION:
        warnings.warn(
            f"imdclient version {imdclient_version} is too old; "
            f"need at least {MIN_IMDCLIENT_VERSION}, Your installed version of "
            "imdclient will NOT be used.",
            category=RuntimeWarning,
        )
        HAS_IMDCLIENT = False

logger = logging.getLogger("MDAnalysis.coordinates.IMDReader")


class IMDReader(StreamReaderBase):
    """
    Coordinate reader implementing the IMDv3 protocol for streaming simulation data.

    This class handles the technical aspects of connecting to IMD-enabled simulation
    engines and processing the incoming data stream. For usage examples and protocol
    overview, see the module documentation above.

    The reader manages socket connections, data buffering, and frame parsing according
    to the IMDv3 specification. It automatically handles different data packet types
    (coordinates, velocities, forces, energies, timing) and populates MDAnalysis
    timestep objects accordingly.

    Parameters
    ----------
    filename : a string of the form "imd://host:port" where host is the hostname
        or IP address of the listening simulation engine's IMD server and port
        is the port number.
    n_atoms : int (optional)
        number of atoms in the system. defaults to number of atoms
        in the topology. Don't set this unless you know what you're doing.
    buffer_size: int (optional) default=10*(1024**2)
        number of bytes of memory to allocate to the :class:`~imdclient.IMDClient.IMDClient`'s
        internal buffer. Defaults to 10 megabytes. Larger buffers can improve
        performance for analyses with periodic heavy computation.
    **kwargs : dict (optional)
        keyword arguments passed to the constructed :class:`~imdclient.IMDClient.IMDClient`

    Notes
    -----
    The IMDReader provides access to additional simulation data through the timestep's
    `data` attribute (`ts.data`). The following keys may be available depending on
    what the simulation engine transmits:

    * `dt` : float
        Time step size in picoseconds (from the `IMD_TIME`_ packet of the IMDv3 protocol)
    * `step` : int
        Current simulation step number (from the `IMD_TIME`_ packet of the IMDv3 protocol)
    * Energy terms : float
        Various energy components (e.g., 'potential', 'kinetic', 'total', etc.)
        from the `IMD_ENERGIES`_ packet of the IMDv3 protocol.

    .. _IMD_TIME: https://imdclient.readthedocs.io/en/latest/protocol_v3.html#time
    .. _IMD_ENERGIES: https://imdclient.readthedocs.io/en/latest/protocol_v3.html#energies

    .. note::
       For important limitations inherent to streaming data, see the module documentation above
       and :class:`~MDAnalysis.coordinates.base.StreamReaderBase` for more technical details.

    .. versionadded:: 2.10.0
    """

    format = "IMD"

    @store_init_arguments
    def __init__(
        self,
        filename,
        n_atoms=None,
        buffer_size=10 * (1024**2),
        **kwargs,
    ):
        if not HAS_IMDCLIENT:
            raise ImportError(
                "IMDReader requires the imdclient package. "
                "Please install it with 'pip install imdclient'."
            )

        super(IMDReader, self).__init__(filename, **kwargs)

        self._imdclient = None
        logger.debug("IMDReader initializing")

        if n_atoms is None:
            raise ValueError("IMDReader: n_atoms must be specified")
        self.n_atoms = n_atoms

        try:
            host, port = parse_host_port(filename)
        except ValueError as e:
            raise ValueError(f"IMDReader: Invalid IMD URL '{filename}': {e}")

        # This starts the simulation
        self._imdclient = IMDClient(
            host, port, n_atoms, buffer_size=buffer_size, **kwargs
        )

        imdsinfo = self._imdclient.get_imdsessioninfo()
        if imdsinfo.version != 3:
            raise ValueError(
                f"IMDReader: Detected IMD version v{imdsinfo.version}, "
                + "but IMDReader is only compatible with v3"
            )

        self.ts = self._Timestep(
            self.n_atoms,
            positions=imdsinfo.positions,
            velocities=imdsinfo.velocities,
            forces=imdsinfo.forces,
            **self._ts_kwargs,
        )

        try:
            self._read_next_timestep()
        except EOFError as e:
            raise RuntimeError(f"IMDReader: Read error: {e}") from e

    def _read_frame(self, frame):

        imdf = self._imdclient.get_imdframe()

        self._frame = frame
        self._load_imdframe_into_ts(imdf)

        logger.debug("IMDReader: Loaded frame %d", self._frame)
        return self.ts

    def _load_imdframe_into_ts(self, imdf):
        self.ts.frame = self._frame
        if imdf.time is not None:
            self.ts.time = imdf.time
            self.ts.data["dt"] = imdf.dt
            self.ts.data["step"] = imdf.step
        if imdf.energies is not None:
            self.ts.data.update(
                {k: v for k, v in imdf.energies.items() if k != "step"}
            )
        if imdf.box is not None:
            self.ts.dimensions = core.triclinic_box(*imdf.box)
        if imdf.positions is not None:
            # must call copy because reference is expected to reset
            # see 'test_frame_collect_all_same' in MDAnalysisTests.coordinates.base
            np.copyto(self.ts.positions, imdf.positions)
        if imdf.velocities is not None:
            np.copyto(self.ts.velocities, imdf.velocities)
        if imdf.forces is not None:
            np.copyto(self.ts.forces, imdf.forces)

    @staticmethod
    def _format_hint(thing):
        if not isinstance(thing, str):
            return False
        # a weaker check for type hint
        if thing.startswith("imd://"):
            return True
        else:
            return False

    def close(self):
        """Gracefully shut down the reader. Stops the producer thread."""
        logger.debug("IMDReader close() called")
        if self._imdclient is not None:
            self._imdclient.stop()
        logger.debug("IMDReader shut down gracefully.")



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

"""ParmEd structure I/O --- :mod:`MDAnalysis.converters.ParmEd`
================================================================

Read coordinates data from a `ParmEd <https://parmed.github.io/ParmEd/html>`_ :class:`parmed.structure.Structure`
with :class:`ParmEdReader` into a MDAnalysis Universe. Convert it back to a
:class:`parmed.structure.Structure` with :class:`ParmEdConverter`.

Example
-------

ParmEd has some neat functions. One such is `HMassRepartition`_.
This function changes the mass of the hydrogens in your system to your desired
value. It then adjusts the mass of the atom to which it is bonded by the same
amount, so that the total mass is unchanged. ::

    >>> import MDAnalysis as mda
    >>> from MDAnalysis.tests.datafiles import PRM
    >>> u = mda.Universe(PRM)
    >>> u.atoms.masses[:10]
    array([14.01 ,  1.008,  1.008,  1.008, 12.01 ,  1.008, 12.01 ,  1.008,
        1.008,  1.008])

We can convert our universe to a ParmEd structure to change our hydrogen
masses. ::

    >>> import parmed.tools as pmt
    >>> parm = u.atoms.convert_to('PARMED')
    >>> hmass = pmt.HMassRepartition(parm, 5)  # convert to 5 daltons
    >>> hmass.execute()

We can then convert it back to an MDAnalysis Universe for further analysis. ::

    >>> u2 = mda.Universe(parm)
    >>> u2.atoms.masses[:10]
    array([2.03399992, 5.        , 5.        , 5.        , 8.01799965,
       5.        , 0.034     , 5.        , 5.        , 5.        ])

.. _`HMassRepartition`: http://parmed.github.io/ParmEd/html/parmed.html#hmassrepartition



Classes
-------

.. autoclass:: ParmEdReader
   :members:

.. autoclass:: ParmEdConverter
   :members:


.. versionchanged:: 2.0.0
   The ParmEdReader and ParmEdConverter classes were moved from :mod:`~MDAnalysis.coordinates`
   to :mod:`~MDAnalysis.converters`

"""
import functools
import itertools
import warnings

from ..guesser.tables import SYMB2Z
import numpy as np

from . import base
from ..coordinates.base import SingleFrameReaderBase
from ..core.universe import Universe
from ..exceptions import NoDataError


class ParmEdReader(SingleFrameReaderBase):
    """Coordinate reader for ParmEd."""

    format = "PARMED"

    # Structure.coordinates always in Angstrom
    units = {"time": None, "length": "Angstrom"}

    @staticmethod
    def _format_hint(thing):
        """Can this reader read *thing*?

        .. versionadded:: 1.0.0
        """
        try:
            import parmed as pmd
        except ImportError:
            # if we can't import parmed, it's probably not parmed
            return False
        else:
            return isinstance(thing, pmd.Structure)

    def _read_first_frame(self):
        self.n_atoms = len(self.filename.atoms)

        self.ts = ts = self._Timestep(self.n_atoms, **self._ts_kwargs)

        if self.filename.coordinates is not None:
            ts._pos = self.filename.coordinates

        # optional field
        ts.dimensions = self.filename.box

        ts.frame = 0
        return ts


MDA2PMD = {
    "tempfactor": "bfactor",
    "gbscreen": "screen",
    "altLoc": "altloc",
    "nbindex": "nb_idx",
    "solventradius": "solvent_radius",
    "id": "number",
}


def get_indices_from_subset(i, atomgroup=None, universe=None):
    return atomgroup[universe.atoms[i]]


class ParmEdConverter(base.ConverterBase):
    """Convert MDAnalysis AtomGroup or Universe to ParmEd :class:`~parmed.structure.Structure`.

    Example
    -------

    .. code-block:: python

        import parmed as pmd
        import MDAnalysis as mda
        from MDAnalysis.tests.datafiles import GRO
        pgro = pmd.load_file(GRO)
        mgro = mda.Universe(pgro)
        parmed_subset = mgro.select_atoms('resname SOL').convert_to('PARMED')


    """

    lib = "PARMED"
    units = {"time": None, "length": "Angstrom"}

    def convert(self, obj):
        """Write selection at current trajectory frame to :class:`~parmed.structure.Structure`.

        Parameters
        -----------
        obj : AtomGroup or Universe or :class:`Timestep`
        """
        try:
            import parmed as pmd
        except ImportError:
            errmsg = (
                "ParmEd is required for ParmEdConverter but is not "
                "installed. Try installing it with \n"
                "pip install parmed"
            )
            raise ImportError(errmsg)
        try:
            # make sure to use atoms (Issue 46)
            ag_or_ts = obj.atoms
        except AttributeError:
            raise TypeError("No atoms found in obj argument") from None

        # Check for topology information
        missing_topology = []
        try:
            names = ag_or_ts.names
        except (AttributeError, NoDataError):
            names = itertools.cycle(("X",))
            missing_topology.append("names")
        try:
            resnames = ag_or_ts.resnames
        except (AttributeError, NoDataError):
            resnames = itertools.cycle(("UNK",))
            missing_topology.append("resnames")

        if missing_topology:
            warnings.warn(
                "Supplied AtomGroup was missing the following attributes: "
                "{miss}. These will be written with default values. "
                "Alternatively these can be supplied as keyword arguments."
                "".format(miss=", ".join(missing_topology))
            )

        try:
            positions = ag_or_ts.positions
        except (AttributeError, NoDataError):
            positions = [None] * ag_or_ts.n_atoms

        try:
            velocities = ag_or_ts.velocities
        except (AttributeError, NoDataError):
            velocities = [None] * ag_or_ts.n_atoms

        atom_kwargs = []
        for atom, name, resname, xyz, vel in zip(
            ag_or_ts, names, resnames, positions, velocities
        ):
            akwargs = {"name": name}
            chain_seg = {"segid": atom.segid}
            for attrname in (
                "mass",
                "charge",
                "type",
                "altLoc",
                "tempfactor",
                "occupancy",
                "gbscreen",
                "solventradius",
                "nbindex",
                "rmin",
                "epsilon",
                "rmin14",
                "epsilon14",
                "id",
            ):
                try:
                    akwargs[MDA2PMD.get(attrname, attrname)] = getattr(
                        atom, attrname
                    )
                except AttributeError:
                    pass
            try:
                el = atom.element.lower().capitalize()
                akwargs["atomic_number"] = SYMB2Z[el]
            except (KeyError, AttributeError):
                try:
                    tp = atom.type.lower().capitalize()
                    akwargs["atomic_number"] = SYMB2Z[tp]
                except (KeyError, AttributeError):
                    pass
            try:
                chain_seg["chain"] = atom.chainID
            except AttributeError:
                pass
            try:
                chain_seg["inscode"] = atom.icode
            except AttributeError:
                pass
            atom_kwargs.append(
                (akwargs, resname, atom.resid, chain_seg, xyz, vel)
            )

        struct = pmd.Structure()

        for akwarg, resname, resid, kw, xyz, vel in atom_kwargs:
            atom = pmd.Atom(**akwarg)
            if xyz is not None:
                atom.xx, atom.xy, atom.xz = xyz

            if vel is not None:
                atom.vx, atom.vy, atom.vz = vel

            atom.atom_type = pmd.AtomType(
                akwarg["name"],
                None,
                akwarg["mass"],
                atomic_number=akwargs.get("atomic_number"),
            )
            struct.add_atom(atom, resname, resid, **kw)

        try:
            struct.box = ag_or_ts.dimensions
        except AttributeError:
            struct.box = None

        if hasattr(ag_or_ts, "universe"):
            atomgroup = {
                atom: index for index, atom in enumerate(list(ag_or_ts))
            }
            get_atom_indices = functools.partial(
                get_indices_from_subset,
                atomgroup=atomgroup,
                universe=ag_or_ts.universe,
            )
        else:
            get_atom_indices = lambda x: x

        # bonds
        try:
            params = ag_or_ts.intra_bonds
        except AttributeError:
            pass
        else:
            for p in params:
                atoms = [
                    struct.atoms[i] for i in map(get_atom_indices, p.indices)
                ]
                try:
                    for obj in p.type:
                        bond = pmd.Bond(*atoms, type=obj.type, order=obj.order)
                        struct.bonds.append(bond)
                    if isinstance(obj.type, pmd.BondType):
                        struct.bond_types.append(bond.type)
                        bond.type.list = struct.bond_types
                except (TypeError, AttributeError):
                    order = p.order if p.order is not None else 1
                    btype = getattr(p.type, "type", None)

                    bond = pmd.Bond(*atoms, type=btype, order=order)
                    struct.bonds.append(bond)
                    if isinstance(bond.type, pmd.BondType):
                        struct.bond_types.append(bond.type)
                        bond.type.list = struct.bond_types

        # dihedrals
        try:
            params = ag_or_ts.dihedrals.atomgroup_intersection(
                ag_or_ts, strict=True
            )
        except AttributeError:
            pass
        else:
            for p in params:
                atoms = [
                    struct.atoms[i] for i in map(get_atom_indices, p.indices)
                ]
                try:
                    for obj in p.type:
                        imp = getattr(obj, "improper", False)
                        ign = getattr(obj, "ignore_end", False)
                        dih = pmd.Dihedral(
                            *atoms, type=obj.type, ignore_end=ign, improper=imp
                        )
                        struct.dihedrals.append(dih)
                        if isinstance(dih.type, pmd.DihedralType):
                            struct.dihedral_types.append(dih.type)
                            dih.type.list = struct.dihedral_types
                except (TypeError, AttributeError):
                    btype = getattr(p.type, "type", None)
                    imp = getattr(p.type, "improper", False)
                    ign = getattr(p.type, "ignore_end", False)
                    dih = pmd.Dihedral(
                        *atoms, type=btype, improper=imp, ignore_end=ign
                    )
                    struct.dihedrals.append(dih)
                    if isinstance(dih.type, pmd.DihedralType):
                        struct.dihedral_types.append(dih.type)
                        dih.type.list = struct.dihedral_types

        for param, pmdtype, trackedlist, typelist, clstype in (
            (
                "ureybradleys",
                pmd.UreyBradley,
                struct.urey_bradleys,
                struct.urey_bradley_types,
                pmd.BondType,
            ),
            (
                "angles",
                pmd.Angle,
                struct.angles,
                struct.angle_types,
                pmd.AngleType,
            ),
            (
                "impropers",
                pmd.Improper,
                struct.impropers,
                struct.improper_types,
                pmd.ImproperType,
            ),
            ("cmaps", pmd.Cmap, struct.cmaps, struct.cmap_types, pmd.CmapType),
        ):
            try:
                params = getattr(ag_or_ts, param)
                values = params.atomgroup_intersection(ag_or_ts, strict=True)
            except AttributeError:
                pass
            else:
                for v in values:
                    atoms = [
                        struct.atoms[i]
                        for i in map(get_atom_indices, v.indices)
                    ]

                    try:
                        for parmed_obj in v.type:
                            p = pmdtype(*atoms, type=parmed_obj.type)
                            trackedlist.append(p)
                            if isinstance(p.type, clstype):
                                typelist.append(p.type)
                                p.type.list = typelist
                    except (TypeError, AttributeError):
                        vtype = getattr(v.type, "type", None)

                        p = pmdtype(*atoms, type=vtype)
                        trackedlist.append(p)
                        if isinstance(p.type, clstype):
                            typelist.append(p.type)
                            p.type.list = typelist
        return struct



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding: utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

"""
GAMESS Topology Parser
======================

.. versionadded:: 0.9.1

Reads a GAMESS_ output file (also Firefly_ and `GAMESS-UK`_) and pulls
element information from it.  Symmetrical assembly is read (not
symmetry element!).  Atom names are read from the GAMESS section.  Any
information about residues or segments will not be populated.

.. _GAMESS: http://www.msg.ameslab.gov/gamess/
.. _Firefly: http://classic.chem.msu.su/gran/gamess/index.html
.. _`GAMESS-UK`: http://www.cfs.dl.ac.uk/


Classes
-------

.. autoclass:: GMSParser
   :members:
   :inherited-members:

"""
import re
import numpy as np

from ..lib.util import openany
from .base import TopologyReaderBase
from ..core.topology import Topology
from ..core.topologyattrs import (
    Atomids,
    Atomnames,
    Resids,
    Resnums,
    Segids,
    AtomAttr,
)


class AtomicCharges(AtomAttr):
    attrname = "atomiccharges"
    singular = "atomiccharge"
    per_object = "atom"


class GMSParser(TopologyReaderBase):
    """GAMESS_ topology parser.

    Creates the following Attributes:
     - names
     - atomic charges


    .. note::

        By default, atomtypes and masses will be guessed on Universe creation.
        This may change in release 3.0.
        See :ref:`Guessers` for more information.

    .. versionadded:: 0.9.1
    .. versionchanged:: 2.8.0
        Removed type and mass guessing (attributes guessing takes place now
        through universe.guess_TopologyAttrs() API).

    """

    format = "GMS"

    def parse(self, **kwargs):
        """Read list of atoms from a GAMESS file."""
        names = []
        at_charges = []

        with openany(self.filename) as inf:
            while True:
                line = inf.readline()
                if not line:
                    raise EOFError
                if re.match(
                    r"^\s+ATOM\s+ATOMIC\s+COORDINATES\s*\(BOHR\).*", line
                ):
                    break
            line = inf.readline()  # skip

            while True:
                line = inf.readline()
                _m = re.match(
                    r"^\s*([A-Za-z_][A-Za-z_0-9]*)\s+([0-9]+\.[0-9]+)\s+(\-?[0-9]+\.[0-9]+)\s+(\-?[0-9]+\.[0-9]+)\s+(\-?[0-9]+\.[0-9]+).*",
                    line,
                )
                if _m is None:
                    break
                name = _m.group(1)
                at_charge = int(float(_m.group(2)))

                names.append(name)
                at_charges.append(at_charge)
                # TODO: may be use coordinates info from _m.group(3-5) ??

        n_atoms = len(names)
        attrs = [
            Atomids(np.arange(n_atoms) + 1),
            Atomnames(np.array(names, dtype=object)),
            AtomicCharges(np.array(at_charges, dtype=np.int32)),
            Resids(np.array([1])),
            Resnums(np.array([1])),
            Segids(np.array(["SYSTEM"], dtype=object)),
        ]
        top = Topology(n_atoms, 1, 1, attrs=attrs)

        return top



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
from itertools import permutations

import numpy as np
import pytest
from numpy.testing import (
    assert_allclose,
    assert_equal,
    assert_almost_equal,
    assert_array_equal,
)

from MDAnalysis.lib import transformations as t

from unittest import TestCase


"""
Testing transformations is weird because there are 2 versions of many of
these functions.  This is because both python and Cython versions of
these functions exist.  To test therefore, each test has to be done twice,
once for each backend.  This is done through parametrizing and passing
in both versions of the function as an argument.

This should ensure that both versions work and are covered!

.. versionchanged:: 1.0.0
   test_transformations_old_module was removed as core/transformations.py is
   gone 
"""

# tolerance for tests
_ATOL = 1e-06


@pytest.mark.parametrize("f", [t._py_identity_matrix, t.identity_matrix])
def test_identity_matrix(f):
    I = f()
    assert_allclose(I, np.dot(I, I))
    assert_equal(np.sum(I), np.trace(I))
    assert_allclose(I, np.identity(4, dtype=np.float64))


@pytest.mark.parametrize(
    "f",
    [
        t._py_translation_matrix,
        t.translation_matrix,
    ],
)
def test_translation_matrix(f):
    v = np.array([0.2, 0.2, 0.2])
    assert_allclose(v, f(v)[:3, 3])


def test_translation_from_matrix():
    # doesn't seem to have a Cython backend
    v0 = np.array([0.2, 0.2, 0.2])
    v1 = t.translation_from_matrix(t.translation_matrix(v0))
    assert_allclose(v0, v1)


@pytest.mark.parametrize(
    "f",
    [
        t._py_reflection_matrix,
        t.reflection_matrix,
    ],
)
def test_reflection_matrix(f):
    v0 = np.array([0.2, 0.2, 0.2, 1.0])  # arbitrary values
    v1 = np.array([0.4, 0.4, 0.4])
    R = f(v0, v1)
    assert_allclose(2.0, np.trace(R))
    assert_allclose(v0, np.dot(R, v0))
    v2 = v0.copy()
    v2[:3] += v1
    v3 = v0.copy()
    v2[:3] -= v1
    assert_allclose(v2, np.dot(R, v3))


def test_reflection_from_matrix():
    v0 = np.array([0.2, 0.2, 0.2])  # arbitrary values
    v1 = np.array([0.4, 0.4, 0.4])
    M0 = t.reflection_matrix(v0, v1)
    point, normal = t.reflection_from_matrix(M0)
    M1 = t.reflection_matrix(point, normal)
    assert_equal(t.is_same_transform(M0, M1), True)


@pytest.mark.parametrize(
    "f",
    [
        t._py_rotation_matrix,
        t.rotation_matrix,
    ],
)
def test_rotation_matrix(f):
    R = f(np.pi / 2.0, [0, 0, 1], [1, 0, 0])
    assert_allclose(np.dot(R, [0, 0, 0, 1]), [1.0, -1.0, 0.0, 1.0])
    angle = 0.2 * 2 * np.pi  # arbitrary value
    direc = np.array([0.2, 0.2, 0.2])
    point = np.array([0.4, 0.4, 0.4])
    R0 = f(angle, direc, point)
    R1 = f(angle - 2 * np.pi, direc, point)
    assert_equal(t.is_same_transform(R0, R1), True)
    R0 = f(angle, direc, point)
    R1 = f(-angle, -direc, point)
    assert_equal(t.is_same_transform(R0, R1), True)
    I = np.identity(4, np.float64)
    assert_allclose(I, f(np.pi * 2, direc), atol=_ATOL)
    assert_allclose(2.0, np.trace(f(np.pi / 2, direc, point)))


def test_rotation_from_matrix():
    angle = 0.2 * 2 * np.pi  # arbitrary values
    direc = np.array([0.2, 0.2, 0.2])
    point = np.array([0.4, 0.4, 0.4])
    R0 = t.rotation_matrix(angle, direc, point)
    angle, direc, point = t.rotation_from_matrix(R0)
    R1 = t.rotation_matrix(angle, direc, point)
    assert_equal(t.is_same_transform(R0, R1), True)


@pytest.mark.parametrize(
    "f",
    [
        t._py_scale_matrix,
        t.scale_matrix,
    ],
)
def test_scale_matrix(f):
    v = np.array([14.1, 15.1, 16.1, 1])
    S = f(-1.234)
    assert_allclose(np.dot(S, v)[:3], -1.234 * v[:3])


def test_scale_from_matrix():
    factor = 7
    origin = np.array([0.2, 0.2, 0.2])  # arbitrary values
    direct = np.array([0.4, 0.4, 0.4])
    S0 = t.scale_matrix(factor, origin)
    factor, origin, direction = t.scale_from_matrix(S0)
    S1 = t.scale_matrix(factor, origin, direction)
    assert_equal(t.is_same_transform(S0, S1), True)
    S0 = t.scale_matrix(factor, origin, direct)
    factor, origin, direction = t.scale_from_matrix(S0)
    S1 = t.scale_matrix(factor, origin, direction)
    assert_equal(t.is_same_transform(S0, S1), True)


@pytest.mark.parametrize(
    "f",
    [
        t._py_projection_matrix,
        t.projection_matrix,
    ],
)
class TestProjectionMatrix(object):
    def test_projection_matrix_1(self, f):
        P = f((0, 0, 0), (1, 0, 0))
        assert_allclose(P[1:, 1:], np.identity(4)[1:, 1:], atol=_ATOL)

    def test_projection_matrix_2(self, f):
        point = np.array([0.2, 0.2, 0.2])  # arbitrary values
        normal = np.array([0.4, 0.4, 0.4])
        direct = np.array([0.6, 0.6, 0.6])
        persp = np.array([0.8, 0.8, 0.8])

        P0 = f(point, normal)
        # TODO: why isn't this used anymore?
        P1 = f(point, normal, direction=direct)
        P2 = f(point, normal, perspective=persp)
        P3 = f(point, normal, perspective=persp, pseudo=True)
        assert_equal(t.is_same_transform(P2, np.dot(P0, P3)), True)

    def test_projection_matrix_3(self, f):
        P = f((3, 0, 0), (1, 1, 0), (1, 0, 0))
        v0 = np.array([14.1, 15.1, 16.1, 1])  # arbitrary values
        v1 = np.dot(P, v0)
        assert_allclose(v1[1], v0[1], atol=_ATOL)
        assert_allclose(v1[0], 3.0 - v1[1], atol=_ATOL)


class TestProjectionFromMatrix(object):
    @staticmethod
    @pytest.fixture()
    def data():
        point = np.array([0.2, 0.2, 0.2])  # arbitrary values
        normal = np.array([0.4, 0.4, 0.4])
        direct = np.array([0.6, 0.6, 0.6])
        persp = np.array([0.8, 0.8, 0.8])
        return point, normal, direct, persp

    def test_projection_from_matrix_1(self, data):
        point, normal, direct, persp = data
        P0 = t.projection_matrix(point, normal)
        result = t.projection_from_matrix(P0)
        P1 = t.projection_matrix(*result)
        assert_equal(t.is_same_transform(P0, P1), True)

    def test_projection_from_matrix_2(self, data):
        point, normal, direct, persp = data
        P0 = t.projection_matrix(point, normal, direct)
        result = t.projection_from_matrix(P0)
        P1 = t.projection_matrix(*result)
        assert_equal(t.is_same_transform(P0, P1), True)

    def test_projection_from_matrix_3(self, data):
        point, normal, direct, persp = data
        P0 = t.projection_matrix(
            point, normal, perspective=persp, pseudo=False
        )
        result = t.projection_from_matrix(P0, pseudo=False)
        P1 = t.projection_matrix(*result)
        assert_equal(t.is_same_transform(P0, P1), True)

    def test_projection_from_matrix_4(self, data):
        point, normal, direct, persp = data
        P0 = t.projection_matrix(point, normal, perspective=persp, pseudo=True)
        result = t.projection_from_matrix(P0, pseudo=True)
        P1 = t.projection_matrix(*result)
        assert_equal(t.is_same_transform(P0, P1), True)


@pytest.mark.parametrize(
    "f",
    [
        t._py_clip_matrix,
        t.clip_matrix,
    ],
)
class TestClipMatrix(object):
    def test_clip_matrix_1(self, f):
        frustrum = np.array([0.1, 0.2, 0.3, 0.4, 0.5, 0.6])  # arbitrary values
        frustrum[1] += frustrum[0]
        frustrum[3] += frustrum[2]
        frustrum[5] += frustrum[4]
        M = f(perspective=False, *frustrum)
        assert_allclose(
            np.dot(M, [frustrum[0], frustrum[2], frustrum[4], 1.0]),
            np.array([-1.0, -1.0, -1.0, 1.0]),
        )
        assert_allclose(
            np.dot(M, [frustrum[1], frustrum[3], frustrum[5], 1.0]),
            np.array([1.0, 1.0, 1.0, 1.0]),
        )

    def test_clip_matrix_2(self, f):
        frustrum = np.array([0.1, 0.2, 0.3, 0.4, 0.5, 0.6])  # arbitrary values
        frustrum[1] += frustrum[0]
        frustrum[3] += frustrum[2]
        frustrum[5] += frustrum[4]
        M = f(perspective=True, *frustrum)
        v = np.dot(M, [frustrum[0], frustrum[2], frustrum[4], 1.0])
        assert_allclose(v / v[3], np.array([-1.0, -1.0, -1.0, 1.0]))
        v = np.dot(M, [frustrum[1], frustrum[3], frustrum[4], 1.0])
        assert_allclose(v / v[3], np.array([1.0, 1.0, -1.0, 1.0]))

    def test_clip_matrix_frustrum_left_right_bounds(self, f):
        """ValueError should be raised if left > right."""
        frustrum = np.array([0.4, 0.3, 0.3, 0.7, 0.5, 1.1])
        with pytest.raises(ValueError):
            f(*frustrum)

    def test_clip_matrix_frustrum_bottom_top_bounds(self, f):
        """ValueError should be raised if bottom > top."""
        frustrum = np.array([0.1, 0.3, 0.71, 0.7, 0.5, 1.1])
        with pytest.raises(ValueError):
            f(*frustrum)

    def test_clip_matrix_frustrum_near_far_bounds(self, f):
        """ValueError should be raised if near > far."""
        frustrum = np.array([0.1, 0.3, 0.3, 0.7, 1.5, 1.1])
        with pytest.raises(ValueError):
            f(*frustrum)


@pytest.mark.parametrize(
    "f",
    [
        t._py_shear_matrix,
        t.shear_matrix,
    ],
)
def test_shear_matrix(f):
    angle = 0.2 * 4 * np.pi  # arbitrary values
    direct = np.array([0.2, 0.2, 0.2])
    point = np.array([0.3, 0.4, 0.5])
    normal = np.cross(direct, np.array([0.8, 0.6, 0.4]))
    S = f(angle, direct, point, normal)
    assert_allclose(1.0, np.linalg.det(S), atol=_ATOL)


def test_shear_from_matrix():
    # This seems to fail sometimes if the random numbers
    # roll certain values....
    # angle = (random.random() - 0.5) * 4*np.pi
    # direct = np.random.random(3) - 0.5
    # point = np.random.random(3) - 0.5
    # normal = np.cross(direct, np.random.random(3))
    # In this random configuration the test will fail about 0.05% of all times.
    # Then we hit some edge-cases of the algorithm. The edge cases for these
    # values are slightly different for the linalg library used (MKL/LAPACK).
    # So here are some of my random numbers
    angle = 2.8969075413405783  # arbitrary values
    direct = np.array([-0.31117458, -0.41769518, -0.01188556])
    point = np.array([-0.0035982, -0.40997482, 0.42241425])
    normal = np.cross(direct, np.array([0.08122421, 0.4747914, 0.19851859]))

    S0 = t.shear_matrix(angle, direct, point, normal)
    angle, direct, point, normal = t.shear_from_matrix(S0)
    S1 = t.shear_matrix(angle, direct, point, normal)
    assert_equal(t.is_same_transform(S0, S1), True)


class TestDecomposeMatrix(object):
    def test_decompose_matrix_1(self):
        T0 = t.translation_matrix((1, 2, 3))
        scale, shear, angles, trans, persp = t.decompose_matrix(T0)
        T1 = t.translation_matrix(trans)
        assert_allclose(T0, T1)

    def test_decompose_matrix_2(self):
        S = t.scale_matrix(0.123)
        scale, shear, angles, trans, persp = t.decompose_matrix(S)
        assert_equal(scale[0], 0.123)

    def test_decompose_matrix_3(self):
        R0 = t.euler_matrix(1, 2, 3)
        scale, shear, angles, trans, persp = t.decompose_matrix(R0)
        R1 = t.euler_matrix(*angles)
        assert_allclose(R0, R1)


def test_compose_matrix():
    scale = np.array([0.2, 0.2, 0.2])  # arbitrary values
    shear = np.array([0.4, 0.4, 0.4])
    angles = np.array([0.6, 0.6, 0.6]) * 2 * np.pi
    trans = np.array([0.8, 0.8, 0.8])
    persp = np.array([0.9, 0.9, 0.9, 0.9])

    M0 = t.compose_matrix(scale, shear, angles, trans, persp)
    result = t.decompose_matrix(M0)
    M1 = t.compose_matrix(*result)
    assert_equal(t.is_same_transform(M0, M1), True)


@pytest.mark.parametrize(
    "f",
    [
        t._py_orthogonalization_matrix,
        t.orthogonalization_matrix,
    ],
)
class TestOrthogonalizationMatrix(object):
    def test_orthogonalization_matrix_1(self, f):
        O = f((10.0, 10.0, 10.0), (90.0, 90.0, 90.0))
        assert_allclose(O[:3, :3], np.identity(3, float) * 10, atol=_ATOL)

    def test_orthogonalization_matrix_2(self, f):
        O = f([9.8, 12.0, 15.5], [87.2, 80.7, 69.7])
        assert_allclose(np.sum(O), 43.063229, atol=_ATOL)


@pytest.mark.parametrize(
    "f",
    [
        t._py_superimposition_matrix,
        t.superimposition_matrix,
    ],
)
def test_superimposition_matrix(f):
    v0 = np.sin(np.linspace(0, 0.99, 30)).reshape(3, 10)  # arbitrary values
    M = f(v0, v0)
    assert_allclose(M, np.identity(4), atol=_ATOL)

    R = t.random_rotation_matrix(np.array([0.3, 0.4, 0.5]))
    v0 = ((1, 0, 0), (0, 1, 0), (0, 0, 1), (1, 1, 1))
    v1 = np.dot(R, v0)
    M = f(v0, v1)
    assert_allclose(v1, np.dot(M, v0), atol=_ATOL)

    v0 = np.sin(np.linspace(-1, 1, 400)).reshape(4, 100)
    v0[3] = 1.0
    v1 = np.dot(R, v0)
    M = f(v0, v1)
    assert_allclose(v1, np.dot(M, v0), atol=_ATOL)

    S = t.scale_matrix(0.45)
    T = t.translation_matrix(np.array([0.2, 0.2, 0.2]) - 0.5)
    M = t.concatenate_matrices(T, R, S)
    v1 = np.dot(M, v0)
    v0[:3] += np.sin(np.linspace(0.0, 1e-9, 300)).reshape(3, -1)
    M = f(v0, v1, scaling=True)
    assert_allclose(v1, np.dot(M, v0), atol=_ATOL)

    M = f(v0, v1, scaling=True, usesvd=False)
    assert_allclose(v1, np.dot(M, v0), atol=_ATOL)

    v = np.empty((4, 100, 3), dtype=np.float64)
    v[:, :, 0] = v0
    M = f(v0, v1, scaling=True, usesvd=False)
    assert_allclose(v1, np.dot(M, v[:, :, 0]), atol=_ATOL)


@pytest.mark.parametrize(
    "f",
    [
        t._py_euler_matrix,
        t.euler_matrix,
    ],
)
class TestEulerMatrix(object):
    def test_euler_matrix_1(self, f):
        R = f(1, 2, 3, "syxz")
        assert_allclose(np.sum(R[0]), -1.34786452)

    def test_euler_matrix_2(self, f):
        R = f(1, 2, 3, (0, 1, 0, 1))
        assert_allclose(np.sum(R[0]), -0.383436184)


@pytest.mark.parametrize(
    "f",
    [
        t._py_euler_from_matrix,
        t.euler_from_matrix,
    ],
)
class TestEulerFromMatrix(object):
    def test_euler_from_matrix_1(self, f):
        R0 = t.euler_matrix(1, 2, 3, "syxz")
        al, be, ga = f(R0, "syxz")
        R1 = t.euler_matrix(al, be, ga, "syxz")
        assert_allclose(R0, R1)

    def test_euler_from_matrix_2(self, f):
        angles = 4.0 * np.pi * np.array([-0.3, -0.3, -0.3])  # arbitrary values
        for axes in t._AXES2TUPLE.keys():
            R0 = t.euler_matrix(axes=axes, *angles)
            R1 = t.euler_matrix(axes=axes, *f(R0, axes))
            assert_allclose(R0, R1, err_msg=("{0} failed".format(axes)))


def test_euler_from_quaternion():
    angles = t.euler_from_quaternion([0.99810947, 0.06146124, 0, 0])
    assert_allclose(angles, [0.123, 0, 0], atol=_ATOL)


@pytest.mark.parametrize(
    "f",
    [
        t._py_quaternion_from_euler,
        t.quaternion_from_euler,
    ],
)
def test_quaternion_from_euler(f):
    q = f(1, 2, 3, "ryxz")
    assert_allclose(q, [0.435953, 0.310622, -0.718287, 0.444435], atol=_ATOL)


@pytest.mark.parametrize(
    "f",
    [
        t._py_quaternion_about_axis,
        t.quaternion_about_axis,
    ],
)
def test_quaternion_about_axis(f):
    q = f(0.123, (1, 0, 0))
    assert_allclose(q, [0.99810947, 0.06146124, 0, 0], atol=_ATOL)


@pytest.mark.parametrize(
    "f",
    [
        t._py_quaternion_matrix,
        t.quaternion_matrix,
    ],
)
class TestQuaternionMatrix(object):
    def test_quaternion_matrix_1(self, f):
        M = f([0.99810947, 0.06146124, 0, 0])
        assert_allclose(M, t.rotation_matrix(0.123, (1, 0, 0)), atol=_ATOL)

    def test_quaternion_matrix_2(self, f):
        M = f([1, 0, 0, 0])
        assert_allclose(M, t.identity_matrix(), atol=_ATOL)

    def test_quaternion_matrix_3(self, f):
        M = f([0, 1, 0, 0])
        assert_allclose(M, np.diag([1, -1, -1, 1]), atol=_ATOL)


@pytest.mark.parametrize(
    "f",
    [
        t._py_quaternion_from_matrix,
        t.quaternion_from_matrix,
    ],
)
class TestQuaternionFromMatrix(object):
    def test_quaternion_from_matrix_1(self, f):
        q = f(t.identity_matrix(), True)
        assert_allclose(q, [1.0, 0.0, 0.0, 0.0], atol=_ATOL)

    def test_quaternion_from_matrix_2(self, f):
        q = f(np.diag([1.0, -1.0, -1.0, 1.0]))
        check = np.allclose(q, [0, 1, 0, 0], atol=_ATOL) or np.allclose(
            q, [0, -1, 0, 0], atol=_ATOL
        )
        assert_equal(check, True)

    def test_quaternion_from_matrix_3(self, f):
        R = t.rotation_matrix(0.123, (1, 2, 3))
        q = f(R, True)
        assert_allclose(
            q, [0.9981095, 0.0164262, 0.0328524, 0.0492786], atol=_ATOL
        )

    def test_quaternion_from_matrix_4(self, f):
        R = [
            [-0.545, 0.797, 0.260, 0],
            [0.733, 0.603, -0.313, 0],
            [-0.407, 0.021, -0.913, 0],
            [0, 0, 0, 1],
        ]
        q = f(R)
        assert_allclose(q, [0.19069, 0.43736, 0.87485, -0.083611], atol=_ATOL)

    def test_quaternion_from_matrix_5(self, f):
        R = [
            [0.395, 0.362, 0.843, 0],
            [-0.626, 0.796, -0.056, 0],
            [-0.677, -0.498, 0.529, 0],
            [0, 0, 0, 1],
        ]
        q = f(R)
        assert_allclose(
            q, [0.82336615, -0.13610694, 0.46344705, -0.29792603], atol=_ATOL
        )

    def test_quaternion_from_matrix_6(self, f):
        R = t.random_rotation_matrix()
        q = f(R)
        assert_equal(t.is_same_transform(R, t.quaternion_matrix(q)), True)


@pytest.mark.parametrize(
    "f",
    [
        t._py_quaternion_multiply,
        t.quaternion_multiply,
    ],
)
def test_quaternion_multiply(f):
    q = f([4, 1, -2, 3], [8, -5, 6, 7])
    assert_allclose(q, [28, -44, -14, 48])


@pytest.mark.parametrize(
    "f",
    [
        t._py_quaternion_conjugate,
        t.quaternion_conjugate,
    ],
)
def test_quaternion_conjugate(f):
    q0 = t.random_quaternion()
    q1 = f(q0)
    check = q1[0] == q0[0] and all(q1[1:] == -q0[1:])
    assert_equal(check, True)


@pytest.mark.parametrize(
    "f",
    [
        t._py_quaternion_inverse,
        t.quaternion_inverse,
    ],
)
def test_quaternion_inverse(f):
    q0 = t.random_quaternion()
    q1 = f(q0)
    assert_allclose(t.quaternion_multiply(q0, q1), [1, 0, 0, 0], atol=_ATOL)


def test_quaternion_real():
    assert_allclose(t.quaternion_real([3.0, 0.0, 1.0, 2.0]), 3.0)


def test_quaternion_imag():
    assert_allclose(t.quaternion_imag([3.0, 0.0, 1.0, 2.0]), [0.0, 1.0, 2.0])


@pytest.mark.parametrize(
    "f",
    [
        t._py_quaternion_slerp,
        t.quaternion_slerp,
    ],
)
def test_quaternion_slerp(f):
    q0 = t.random_quaternion()
    q1 = t.random_quaternion()
    q = f(q0, q1, 0.0)
    assert_allclose(q, q0, atol=_ATOL)

    q = f(q0, q1, 1.0, 1)
    assert_allclose(q, q1, atol=_ATOL)

    q = f(q0, q1, 0.5)
    angle = np.arccos(np.dot(q0, q))

    check = np.allclose(2.0, np.arccos(np.dot(q0, q1)) / angle) or np.allclose(
        2.0, np.arccos(-np.dot(q0, q1)) / angle
    )

    assert_equal(check, True)


@pytest.mark.parametrize(
    "f",
    [
        t._py_random_quaternion,
        t.random_quaternion,
    ],
)
class TestRandomQuaternion(object):
    def test_random_quaternion_1(self, f):
        q = f()
        assert_allclose(1.0, t.vector_norm(q))

    def test_random_quaternion_2(self, f):
        q = f(np.array([0.2, 0.2, 0.2]))
        assert_equal(len(q.shape), 1)
        assert_equal(q.shape[0] == 4, True)


@pytest.mark.parametrize(
    "f",
    [
        t._py_random_rotation_matrix,
        t.random_rotation_matrix,
    ],
)
def test_random_rotation_matrix(f):
    R = f()
    assert_allclose(np.dot(R.T, R), np.identity(4), atol=_ATOL)


@pytest.mark.parametrize(
    "f",
    [
        t._py_inverse_matrix,
        t.inverse_matrix,
    ],
)
class TestInverseMatrix(object):
    @pytest.mark.parametrize("size", list(range(1, 7)))
    def test_inverse(self, size, f):
        # Create a known random state to generate numbers from
        # these numbers will then be uncorrelated but deterministic
        rs = np.random.RandomState(1234)
        M0 = rs.randn(size, size)
        M1 = f(M0)
        assert_allclose(M1, np.linalg.inv(M0), err_msg=str(size), atol=_ATOL)

    def test_inverse_matrix(self, f):
        M0 = t.random_rotation_matrix()
        M1 = f(M0.T)
        assert_allclose(M1, np.linalg.inv(M0.T))


@pytest.mark.parametrize(
    "f",
    [
        t._py_is_same_transform,
        t.is_same_transform,
    ],
)
class TestIsSameTransform(object):
    def test_is_same_transform_1(self, f):
        assert_equal(f(np.identity(4), np.identity(4)), True)

    def test_is_same_transform_2(self, f):
        assert_equal(f(t.random_rotation_matrix(), np.identity(4)), False)


@pytest.mark.parametrize(
    "f",
    [
        t._py_random_vector,
        t.random_vector,
    ],
)
class TestRandomVector(object):
    def test_random_vector_1(self, f):
        v = f(1000)
        check = np.all(v >= 0.0) and np.all(v < 1.0)
        assert_equal(check, True)

    def test_random_vector_2(self, f):
        v0 = f(10)
        v1 = f(10)
        assert_equal(np.any(v0 == v1), False)


@pytest.mark.parametrize(
    "f",
    [
        t._py_unit_vector,
        t.unit_vector,
    ],
)
class TestUnitVector(object):
    def test_unit_vector_1(self, f):
        v0 = np.array([0.2, 0.2, 0.2])
        v1 = f(v0)
        assert_allclose(v1, v0 / np.linalg.norm(v0), atol=_ATOL)

    def test_unit_vector_2(self, f):
        v0 = np.sin(np.linspace(0, 10, 5 * 4 * 3)).reshape(5, 4, 3)
        v1 = f(v0, axis=-1)
        v2 = v0 / np.expand_dims(np.sqrt(np.sum(v0 * v0, axis=2)), 2)
        assert_allclose(v1, v2, atol=_ATOL)

    def test_unit_vector_3(self, f):
        v0 = np.sin(np.linspace(0, 10, 5 * 4 * 3)).reshape(5, 4, 3)
        v1 = f(v0, axis=1)
        v2 = v0 / np.expand_dims(np.sqrt(np.sum(v0 * v0, axis=1)), 1)
        assert_allclose(v1, v2, atol=_ATOL)

    def test_unit_vector_4(self, f):
        v0 = np.sin(np.linspace(0, 10, 5 * 4 * 3)).reshape(5, 4, 3)
        v1 = np.empty((5, 4, 3), dtype=np.float64)
        v2 = v0 / np.expand_dims(np.sqrt(np.sum(v0 * v0, axis=1)), 1)
        f(v0, axis=1, out=v1)
        assert_allclose(v1, v2, atol=_ATOL)

    def test_unit_vector_5(self, f):
        assert_equal(list(f([])), [])

    def test_unit_vector_6(self, f):
        assert_equal(list(f([1.0])), [1.0])


@pytest.mark.parametrize(
    "f",
    [
        t._py_vector_norm,
        t.vector_norm,
    ],
)
class TestVectorNorm(object):
    def test_vector_norm_1(self, f):
        v = np.array([0.2, 0.2, 0.2])
        n = f(v)
        assert_allclose(n, np.linalg.norm(v), atol=_ATOL)

    def test_vector_norm_2(self, f):
        v = np.sin(np.linspace(0, 10, 6 * 5 * 3)).reshape(6, 5, 3)
        n = f(v, axis=-1)
        assert_allclose(n, np.sqrt(np.sum(v * v, axis=2)), atol=_ATOL)

    def test_vector_norm_3(self, f):
        v = np.sin(np.linspace(0, 10, 6 * 5 * 3)).reshape(6, 5, 3)
        n = f(v, axis=1)
        assert_allclose(n, np.sqrt(np.sum(v * v, axis=1)), atol=_ATOL)

    def test_vector_norm_4(self, f):
        v = np.sin(np.linspace(0, 10, 5 * 4 * 3)).reshape(5, 4, 3)
        n = np.empty((5, 3), dtype=np.float64)
        f(v, axis=1, out=n)
        assert_allclose(n, np.sqrt(np.sum(v * v, axis=1)), atol=_ATOL)

    def test_vector_norm_5(self, f):
        assert_equal(f([]), 0.0)

    def test_vector_norm_6(self, f):
        assert_equal(f([1.0]), 1.0)


class TestArcBall(object):
    def test_arcball_1(self):
        ball = t.Arcball()
        ball = t.Arcball(initial=np.identity(4))
        ball.place([320, 320], 320)
        ball.down([500, 250])
        ball.drag([475, 275])
        R = ball.matrix()
        assert_allclose(np.sum(R), 3.90583455, atol=_ATOL)

    def test_arcball_2(self):
        ball = t.Arcball(initial=[1, 0, 0, 0])
        ball.place([320, 320], 320)
        ball.setaxes([1, 1, 0], [-1, 1, 0])
        ball.setconstrain(True)
        ball.down([400, 200])
        ball.drag([200, 400])
        R = ball.matrix()
        assert_allclose(np.sum(R), 0.2055924)


def test_rotaxis_equal_vectors():
    a = np.arange(3)
    x = t.rotaxis(a, a)
    assert_array_equal(x, [1, 0, 0])


def test_rotaxis_different_vectors():
    # use random coordinate system
    e = np.eye(3)
    r = np.array(
        [
            [0.69884766, 0.59804425, -0.39237102],
            [0.18784672, 0.37585347, 0.90744023],
            [0.69016342, -0.7078681, 0.15032367],
        ]
    )
    re = np.dot(r, e)

    for i, j, l in permutations(range(3)):
        x = t.rotaxis(re[i], re[j])
        # use abs since direction doesn't matter
        assert_almost_equal(np.abs(np.dot(x, re[l])), 1)
